# -*- coding: utf-8 -*-
# 게임 진행 소스
#
#   This file is part of otak-godori.
#
#   Copyright (C) 2007 Otak-godori Project.
#
#   This program is free software; you can redistribute it and/or modify it
#   under the terms of the GNU General Public License as published by the
#   Free Software Foundation; either version 3, or (at your option) any later
#   version.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#   Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  

import random
import copy
import time
import threading
import gobject
import gtk
import otak

from sound.ogsound import *

from widget.ogcard import *
from ogconf import *
from ogplayuidef import *

# 화투장을 바닥패에 낼때 타입
(
    SUBMIT_TYPE_NONE, 
    SUBMIT_TYPE_PLYR, 
    SUBMIT_TYPE_OPEN
) = range(3)


class OgPlayState:

    def __init__(self):

        self.card_id_center = None
        self.card_id_player = None
        self.highlight_card = None
        self.player_id = 0
        self.match_spread_ids = []
        self.submit_lock = True
        self.submit_type = SUBMIT_TYPE_NONE
        self.submit_card_id = None
        self.submit_empty_card_id = None


    def get_card_id_center(self):
        return self.card_id_center

    def set_card_id_center(self, card_id_center):
        self.card_id_center = card_id_center


    def get_card_id_player(self):
        return self.card_id_player

    def set_card_id_player(self, card_id_player):
        self.card_id_player = card_id_player


    def get_highlight_card(self):
        return self.highlight_card

    def set_highlight_card(self, highlight_card):
        self.highlight_card = highlight_card


    def get_player_id(self):
        return self.player_id
    
    def set_player_id(self, player_id):
        self.player_id = player_id


    def get_match_spread_ids(self):
        return self.match_spread_ids

    def append_match_spread_ids(self, value):
        self.match_spread_ids.append(value)

    def clear_match_spread_ids(self):
        self.match_spread_ids = []


    def is_submit_lock(self):
        return self.submit_lock

    def set_submit_lock(self, submit_lock):
        self.submit_lock = submit_lock


    def get_submit_type(self):
        return self.submit_type
    
    def set_submit_type(self, submit_type):
        self.submit_type = submit_type


    def get_submit_card_id(self):
        return self.submit_card_id

    def set_submit_card_id(self, id):
        self.submit_card_id = id


    def get_submit_empty_card_id(self):
        return self.submit_empty_card_id

    def set_submit_empty_card_id(self, id):
        self.submit_empty_card_id = id


    def clear_submit(self):

        self.submit_match_spread_ids = []
        self.submit_type = SUBMIT_TYPE_NONE
        self.submit_card_id = None
        self.submit_empty_card_id = None


class OgPlaySpread:

    def __init__(self):

        self.clear()


    def clear(self):

        self.cards = []
        self.joker_cnt = 0


    def get_card(self, index):
        return self.cards[index]

    def get_card_cnt(self):
        return len(self.cards)


    def get_joker_cnt(self):
        return self.joker_cnt


    def get_real_cnt(self):
        return self.get_card_cnt() - self.get_joker_cnt()


    def append(self, card):

        self.cards.append(card)

        if card.type == OG_CARD_TYPE_JOKER:
            self.joker_cnt += 1


    def remove(self, card_id = -1):

        card = self.cards.pop(card_id)

        if card.type == OG_CARD_TYPE_JOKER:
            self.joker_cnt -= 1

        return card


class OgPlayPoint:

    def __init__(self):

        self.light_cards = []
        self.ten_cards = []
        self.belt_cards = []
        self.cover_cards = []

        self.go_step = 0
        self.go_point = 0


    def append_light_cards(self, card):
        self.light_cards.append(card)

    def get_light_cards_cnt(self):
        return len(self.light_cards)


    def append_ten_cards(self, card):
        self.ten_cards.append(card)

    def get_ten_cards_cnt(self):
        return len(self.ten_cards)


    def append_belt_cards(self, card):
        self.belt_cards.append(card)

    def get_belt_cards_cnt(self):
        return len(self.belt_cards)


    def append_cover_cards(self, card):
        self.cover_cards.append(card)

    def get_cover_cards_cnt(self):
        return len(self.cover_cards)


    def get_go_step(self):
        return self.go_step

    def next_go_step(self):

        self.set_go_point(self.get_total_point())
        self.go_step += 1


    def get_go_point(self):
        return self.go_point

    def set_go_point(self, point):
        self.go_point = point


    def get_sub_type_cnt(self, cards, sub_type):

        cnt = 0

        for card in cards:
            if card.sub_type == sub_type:
                cnt += 1

        return cnt


    def get_bird_cnt(self):
        return self.get_sub_type_cnt(self.ten_cards, OG_CARD_SUB_TYPE_BIRD)


    def get_blue_cnt(self):
        return self.get_sub_type_cnt(self.belt_cards, OG_CARD_SUB_TYPE_BIRD)

    def get_red_cnt(self):
        return self.get_sub_type_cnt(self.belt_cards, OG_CARD_SUB_TYPE_RED)

    def get_grass_cnt(self):
        return self.get_sub_type_cnt(self.belt_cards, OG_CARD_SUB_TYPE_GRASS)


    def get_double_cnt(self):
        return self.get_sub_type_cnt(self.cover_cards, 
                                     OG_CARD_SUB_TYPE_DOUBLE)

    def get_triple_cnt(self):
        return self.get_sub_type_cnt(self.cover_cards, 
                                     OG_CARD_SUB_TYPE_TRIPLE)


    def get_light_point(self):

        pnt = 0
        rain = False

        cnt = self.get_light_cards_cnt()

        for i in range(cnt):

            if self.light_cards[i].number == CARD_RAIN_NUM:
                rain = True
                break

        if cnt >= 3:

            if cnt == 3:

                if rain:
                    pnt = 2
                else:
                    pnt = 3

            elif cnt == 5:
                pnt = 15

            else:
                pnt = cnt

        return pnt


    def get_ten_point(self):

        pnt = 0

        cnt = self.get_ten_cards_cnt()

        if cnt >= 5:
            pnt = cnt - 4

        return pnt


    def get_belt_point(self):

        pnt = 0

        cnt = self.get_belt_cards_cnt()

        if cnt >= 5:
            pnt = cnt - 4

        return pnt


    def get_cover_point(self):

        pnt = 0

        cnt = self.get_cover_cards_cnt() + self.get_double_cnt() \
                + (self.get_triple_cnt() * 2)

        if cnt >= 10:
            pnt = cnt - 9

        return pnt


    def get_bird_point(self):

        pnt = 0

        cnt = self.get_bird_cnt()

        if cnt == 3:
            pnt = 5

        return pnt


    def get_blue_point(self):

        pnt = 0

        cnt = self.get_blue_cnt()

        if cnt == 3:
            pnt = 3

        return pnt


    def get_red_point(self):

        pnt = 0

        cnt = self.get_red_cnt()

        if cnt == 3:
            pnt = 3

        return pnt


    def get_grass_point(self):

        pnt = 0

        cnt = self.get_grass_cnt()

        if cnt == 3:
            pnt = 3

        return pnt


    def get_total_point(self):

        pnt = 0

        pnt += self.get_light_point()
        pnt += self.get_ten_point()
        pnt += self.get_belt_point()
        pnt += self.get_cover_point()
        pnt += self.get_bird_point()
        pnt += self.get_blue_point()
        pnt += self.get_red_point()
        pnt += self.get_grass_point()

        return pnt


    def get_final_point(self):

        pnt = self.get_total_point()

        # 멍따
        if self.get_ten_cards_cnt() >= 7:
            pnt *= 2

        return pnt


class OgPlayCardChoice(gtk.Window):

    def __init__(self, parent):

        gtk.Window.__init__(self, gtk.WINDOW_POPUP)

        self.response_id = None
        self.loop = None

        self.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        self.set_transient_for(parent)

        fixed = gtk.Fixed()

        self.set_transient_for(parent)
        self.set_default_size(300, 250)
        self.set_modal(True)

        self.event1 = gtk.EventBox()
        self.image1 = gtk.Image()
        self.event1.add(self.image1)
        self.event1.set_size_request(100, 155)

        self.event2 = gtk.EventBox()
        self.image2 = gtk.Image()
        self.event2.add(self.image2)
        self.event2.set_size_request(100, 155)
        
        fixed.put(self.event1, 30, 30)
        fixed.put(self.event2, 170, 30)
        fixed.show_all()

        self.add(fixed)

        self.event1.connect("button_press_event",
                            self.on_card_1_button_press_event, None)
        self.event2.connect("button_press_event",
                            self.on_card_2_button_press_event, None)


    def shutdown_loop(self):

        if self.loop:
            if self.loop.is_running():
                self.loop.quit()


    def on_card_1_button_press_event(self, widget, event, user_data):

        self.response_id = 0
		
        self.shutdown_loop()

        return False


    def on_card_2_button_press_event(self, widget, event, user_data):

        self.response_id = 1

        self.shutdown_loop()

        return False


    def run(self, pixbuf1, pixbuf2):

        sndplay = OgSound(2)

        self.image1.set_from_pixbuf(pixbuf1)
        self.image2.set_from_pixbuf(pixbuf2)

        self.show()

        self.response_id = None
        self.loop = gobject.MainLoop()

        gtk.gdk.threads_leave()
        self.loop.run()
        gtk.gdk.threads_enter()

        self.hide()

        return self.response_id


class OgPlayGoChoice(gtk.Window):

    def __init__(self, parent):

        gtk.Window.__init__(self, gtk.WINDOW_POPUP)

        self.response_is_go = None
        self.loop = None

        self.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        self.set_transient_for(parent)

        fixed = gtk.Fixed()

        self.set_transient_for(parent)
        self.set_default_size(300, 250)
        self.set_modal(True)

        filepath = "%s/button/go.png" % IMG_PREFIX
        self.button = otak.wdg.ImageButton(filepath, 90, 50)
        self.button.connect("button_press_event",
                            self.on_go_button_press_event, None)
        fixed.put(self.button, 30, 170)

        filepath = "%s/button/stop.png" % IMG_PREFIX
        self.button = otak.wdg.ImageButton(filepath, 90, 50)
        self.button.connect("button_press_event",
                            self.on_stop_button_press_event, None)
        fixed.put(self.button, 180, 170)

        fixed.show_all()

        self.add(fixed)


    def shutdown_loop(self):

        if self.loop:
            if self.loop.is_running():
                self.loop.quit()


    def on_go_button_press_event(self, widget, event, user_data):

        self.response_is_go = True

        self.shutdown_loop()

        return False


    def on_stop_button_press_event(self, widget, event, user_data):

        self.response_is_go = False

        self.shutdown_loop()

        return False


    def run(self):

        self.show()

        self.response_is_go = None
        self.loop = gobject.MainLoop()

        gtk.gdk.threads_leave()
        self.loop.run()
        gtk.gdk.threads_enter()

        self.hide()

        return self.response_is_go


class OgPlayUi(gtk.Window):

    def __init__(self, parent=None):

        gtk.Window.__init__(self)

        self.eventbox = gtk.EventBox()
        self.eventbox.show()
        self.fixed = gtk.Fixed()
        self.fixed.show()

        self.eventbox.add(self.fixed)
        self.add(self.eventbox)

        self.set_default_size(800, 600)

        self.eventbox.connect("enter_notify_event", 
                            self.on_fixed_enter_notify_event, None)

        # 네트웍 플레이 여부 (True이면 상대 숨은패 정보를 가지지 않는다)
        self.network_play = False
        # 바닥패 공간 정보 (늘었다 줄었다하는것이 아니라 공간 수는 고정이다)
        self.spread_spaces = []

        for i in range(SPRD_CNT):
            self.spread_spaces.append(OgPlaySpread())

        self.construct_cards_full()
        self.construct_cards_empty()

        self.card_choice_window = OgPlayCardChoice(self)
        self.go_choice_window = OgPlayGoChoice(self)


    # 전체 화투장 생성
    def construct_cards_full(self):

        # 전체 화투장 (섞이지 않은 순수 화투장 객체 리스트)
        self.cards_full = []

        for i in range(CARD_CNT):

            num      = int(i / 4) + 1
            sub_num  = (i % 4) + 1
            filepath = "%s/card/%02d_%02d.jpg" % (IMG_PREFIX, num, sub_num)
            card     = OgCard()

            card.set_from_file(filepath)
            card.set_number(num)
            card.set_type(CARD_TYPES[i][0], CARD_TYPES[i][1])
            card.set_size_request(CARD_NML_WIDTH, CARD_NML_HEIGHT)
            card.hide()

            self.cards_full.append(card)

    # 빈 화투장 생성
    def construct_cards_empty(self):

        # 빈 화투장 (화면에 뒤집혀 보인며 화투장의 속성 정보가 없다.)
        self.cards_empty = []

        pixbuf = gtk.gdk.pixbuf_new_from_file(IMG_PREFIX + "/card/hide.jpg")
        pixbuf = pixbuf.scale_simple(CARD_NML_WIDTH, CARD_NML_HEIGHT, 
                                     gtk.gdk.INTERP_BILINEAR)

        for i in range(CARD_CNT):

            card = OgCard()

            card.set_size_request(CARD_NML_WIDTH, CARD_NML_HEIGHT)
            card.set_cover_pixbuf(pixbuf)
            card.cover_show()
            card.show()
            self.cards_empty.append(card)


    # signal callback. 바탕에 마우스 포인터가 진입하면 화면 정리를 한다.
    def on_fixed_enter_notify_event(self, widget, event, user_data):

        # 강조된 화투장의 강종표시를 없엔다.
        self.card_highlight_cb(None)

        return False


    # signal_callback. 화투장을 마우스로 누르면 제출한다.
    def on_player_card_button_press_event(self, widget, event, 
                                          user_data):

        if not self.state_bean.is_submit_lock():

            self.state_bean.set_submit_lock(True)

            card_id = self.get_card_id_by_card(widget)
            self.submit_player_card(card_id)

        return False


    def thread_action(self, action_cb, interval):

        timer = threading.Timer(interval, action_cb)
        timer.start()


    # 화투장에 마우스 포인터가 진입할때의 강조기능에 함께 실행할 callback
    def card_highlight_cb(self, card):

        hightlight_card = self.state_bean.get_highlight_card()

        if hightlight_card:
            hightlight_card.highlight_off()

        self.state_bean.set_highlight_card(card)


    # 화투장을 마우스로 누를수 있는 기능 활성화
    def card_click_on(self, card):

        card.highlight_connect(self.card_highlight_cb)
        card.connect("button_press_event", 
                     self.on_player_card_button_press_event, None)


    # 화투장을 마우스로 누를수 있는 기능 비활성화
    def card_click_off(self, card):

        card.highlight_disconnect()
        card.highlight_off()
        card.disconnect_by_func(self.on_player_card_button_press_event)


    # 화투장을 이동한다.
    def move_card(self, card, x, y, resent):

        if resent:
            self.fixed.remove(card)
            self.fixed.put(card, x, y)

        else:
            self.fixed.move(card, x, y)

        card.show()


    # 바닥패에 화투장을 추가한다.
    def append_spread_card(self, spread_id, card):

        spread = self.spread_spaces[spread_id]

        # spread객체에 아이템 추가
        spread.append(card)

        # 위젯 이동
        cnt   = spread.get_card_cnt()
        space = (cnt - 1) * CARD_SPC_OVER
        x     = SPRD_POSITION[spread_id][0] + space
        y     = SPRD_POSITION[spread_id][1] + space

        self.move_card(card, x, y, True)


    # 바닥패를 병합한다.
    def merge_spreads(self, src_spread_id, dest_spread_id):

        src_spread  = self.spread_spaces[src_spread_id]
        dest_spread = self.spread_spaces[dest_spread_id]

        while src_spread.cards:

            card = src_spread.remove()
            self.append_spread_card(dest_spread_id, card)


    def get_spreads_using_cnt(self):

        cnt = 0

        for space in self.spread_spaces:

            card_cnt = space.get_card_cnt()

            if card_cnt > 0:
                cnt += 1

        return cnt


    # 바닥패중 화투장의 번호가 card_num과 같은 것들의 정보를 map으로 반환
    def get_spread_infos(self, card_num):

        dic = {KEY_SPRD_MTCH_CNT:0}

        for i in range(SPRD_CNT):

            spread = self.spread_spaces[i]

            if spread.cards:

                # 화투장의 기본정보를 가져오기위해 첫번째 화투장을 이용.
                # 조커는 첫번째 화투장일수 없음
                card = spread.cards[0]

                if card.number == card_num:

                    real_cnt = spread.get_real_cnt()
                    dic[KEY_SPRD_MTCH_CNT] += 1

                    match_cnt = dic[KEY_SPRD_MTCH_CNT]
                    if match_cnt == 1:
                        otak.util.Dic.setItem(dic, KEY_SPRD_FRST_ID, i,
                                        ignr_has = True)
                    elif match_cnt >= 2:
                        otak.util.Dic.setItem(dic, KEY_SPRD_SCND_ID, i,
                                        ignr_has = True)

                    if real_cnt == 1:
                        otak.util.Dic.setItem(dic, KEY_SPRD_SIGL_ID, i, 
                                        ignr_has = True)
                    elif real_cnt > 1:
                        otak.util.Dic.setItem(dic, KEY_SPRD_MLTI_ID, i, 
                                        ignr_has = True)

            else:
                otak.util.Dic.setItem(dic, KEY_SPRD_EMPT_ID, i,
                                      ignr_has = True)

        return dic


    def card_choice_show(self, spread_id_1, spread_id_2):

        spread1 = self.spread_spaces[spread_id_1]
        spread2 = self.spread_spaces[spread_id_2]
        card1   = spread1.get_card(0)
        card2   = spread2.get_card(0)

        return self.card_choice_window.run(card1.pixbuf, card2.pixbuf)


    # 플레이어의 화투장을 제출한다.
    def submit_player_card(self, card_id):

        self.state_bean.clear_submit()

        self.state_bean.set_submit_type(SUBMIT_TYPE_PLYR)
        self.state_bean.set_submit_card_id(card_id)

        player_id = self.state_bean.get_player_id()

        if player_id != MY_PLYR_ID:
            empty_card_id = self.player_hidden_empty_card_ids[player_id].pop()
            self.state_bean.set_submit_empty_card_id(empty_card_id)

        self.thread_action(self.submit, 0)


    def submit_player_computer(self):

        # TODO : sohalee 테스트를 위해 마지막 카드를 낸다.
        # 추후에 점수패와 가진패를 참고하여 우선순위 알고리즘을 만들자
        player_id = self.state_bean.get_player_id()
        card_id = self.player_hidden_card_ids[player_id].pop()

        self.submit_player_card(card_id)


    def get_spread_id_by_infos(self, infos):

        match_cnt = infos[KEY_SPRD_MTCH_CNT]
        spread_id = -1

        if infos.has_key(KEY_SPRD_EMPT_ID):
            spread_id = infos[KEY_SPRD_EMPT_ID]

        if match_cnt > 1:

            # 둘중 하나가 두장이 겹쳐있고 다른 하나가 한장이 경우이다.
            if infos.has_key(KEY_SPRD_MLTI_ID):
                spread_id = infos[KEY_SPRD_SIGL_ID]

            # 한공간에 두장이상 겹친 (조커 빼고) 공간이 하나도 없는 것이다.
            # 둘중하나를 선택하게 해주어야한다.
            else:
                ids = [infos[KEY_SPRD_FRST_ID], infos[KEY_SPRD_SCND_ID]]
                id = self.card_choice_show(ids[0], ids[1])
                spread_id = ids[id]

        elif match_cnt == 1:

            # 일치하는 한공간이 두장이상 (조커 빼고)  겹쳐져 있는 경우. 쌌다
            if infos.has_key(KEY_SPRD_MLTI_ID):
                spread_id = infos[KEY_SPRD_MLTI_ID]

            # 일치하는 한공간이 한장만 있는경우 그곳에 넣으면 된다
            else:
                spread_id = infos[KEY_SPRD_SIGL_ID]
            # 그밖의 경우에는 그냥 버리면 된다. spread_id 는 빈공간의 index

        return spread_id


    # 화투장을 제출한다.
    def submit(self):

        snd_tick = OgSound(1)

        card_id = self.state_bean.get_submit_card_id()
        empty_card_id = self.state_bean.get_submit_empty_card_id()
        card = self.cards_full[card_id]
        infos = self.get_spread_infos(card.number)

        gtk.gdk.threads_enter()

        self.card_click_off(card)

        if empty_card_id != None:
            self.cards_empty[empty_card_id].hide()

        spread_id = self.get_spread_id_by_infos(infos)
        self.state_bean.append_match_spread_ids(spread_id)

        self.append_spread_card(spread_id, card)

        gtk.gdk.threads_leave()


        if self.state_bean.get_submit_type() == SUBMIT_TYPE_PLYR:
            time.sleep(0.5)
            self.submit_center_card()

        else:
            time.sleep(1)
            self.submit_pickup()


    def submit_center_card(self):

        self.state_bean.clear_submit()

        self.state_bean.set_submit_type(SUBMIT_TYPE_OPEN)

        card_id = self.get_curr_card_id_center()
        empty_card_id = self.get_curr_empty_card_id_center()
        self.state_bean.set_submit_card_id(card_id)
        self.state_bean.set_submit_empty_card_id(empty_card_id)

        self.card_ids_center.pop()

        # 현재 메소드 내부가 thread에서 호출된 것이므로 submit은 그냥 호출
        self.submit()


    def submit_is_go(self):

        go = True

        player_id = self.state_bean.get_player_id()
        point = self.player_points[player_id]

        total_pnt = point.get_total_point()

        if total_pnt >= VALID_PNT and total_pnt > point.get_go_point():

            gtk.gdk.threads_enter()
            go = self.go_choice_window.run()
            gtk.gdk.threads_leave()

            point.next_go_step()

        return go


    #
    # 제출된 패들과 바닥패를 확인하여 점수패로 가져온다.
    # 반드시 thread로 동작시켜야 한다.
    #
    def submit_pickup(self):

        def submit_pickup_spread(self, spread):
    
            player_id = self.state_bean.get_player_id()
    
            real_cnt = spread.get_real_cnt()
            normal = False
            other = False
    
            # 조커만 먹는다
            if real_cnt == 1:
                if spread.get_joker_cnt() < 1:
                    return other
            # 다 먹는다
            elif real_cnt == 2:
                normal = True
            # 쌌다 못먹는다. 조커도 그냥 같이 덮는다. 먹은 사람이 다가져간다.
            elif real_cnt == 3:
                return other
            # 다 먹고 상대편 점수패중 피(껍데기)를 하나 가져온다.
            # 1피부터 가져오고 없으면 쌍피
            elif real_cnt == 4:
                normal = True
                other  = True
    
            # 뭔가 먹을게 있을때만 이리로 온다. 없으면 위에서 걸러진다 
    
            for j in range(spread.get_card_cnt()):
    
                card = spread.get_card(j)
    
                if not normal and card.type != OG_CARD_TYPE_JOKER:
                    continue
    
                gtk.gdk.threads_enter()
                self.append_player_point_card(player_id, card)
                gtk.gdk.threads_leave()
                time.sleep(0.1)
    
            spread.clear()
    
            return other

    
        player_id = self.state_bean.get_player_id()
        other = False

        # 1. 본인이 내거나 뒤집은 화투장이 있는 공간만을 처리하는 루프이다.
        # 2. 조커외에 세장이면 싼거다. (1번의 조건에 따라)
        # 3. 반드시 조커이외의 화투장이 한장이상 있는 상태이다.
        # 4. 바닥에 조커가 있는 경우는 뒤집은 경우 뿐임.
        #          (자기것은 그냥 바로 점수패 공간으로 넘긴다.
        #          조커를 뒤집으면 자기가 낸 패 공간에 넣으므로 별도의
        #          바닥 공간을 사용하지 않는다.)
        for i in self.state_bean.get_match_spread_ids():

            spread = self.spread_spaces[i]

            if submit_pickup_spread(self, spread):
                other = True

        # 바닥패 쓸이의 경우. 상대편들 점수패중 피(껍데기)를 가져온다.
        # 1피부터 가져오고 없으면 쌍피
        if self.get_spreads_using_cnt() < 1:
            other = True

        if other:
            # TODO : sohalee 상대편 패 가져올 패를 검사하여 가져온다.
            pass


        if self.submit_is_go():

            next_player_id = self.next_player()
    
            if next_player_id == MY_PLYR_ID:
                self.state_bean.set_submit_lock(False)
    
            elif not self.network_play:
                time.sleep(2)
                self.submit_player_computer()


    def append_player_point_card(self, player_id, card):

        pnt = self.player_points[player_id]

        x   = -1
        y   = -1

        if card.type == OG_CARD_TYPE_LIGHT:

            cnt = pnt.get_light_cards_cnt()
            x   = CARD_PSTN_PLYR_PNT_LGHT_X (player_id, cnt)
            y   = CARD_PSTN_PLYR_PNT_LGHT_Y (player_id)

            pnt.append_light_cards(card)

        elif card.type == OG_CARD_TYPE_TEN:

            cnt = pnt.get_ten_cards_cnt()
            x   = CARD_PSTN_PLYR_PNT_TEN_X (player_id, cnt)
            y   = CARD_PSTN_PLYR_PNT_TEN_Y (player_id)

            pnt.append_ten_cards(card)

        elif card.type == OG_CARD_TYPE_BELT:

            cnt = pnt.get_belt_cards_cnt()
            x   = CARD_PSTN_PLYR_PNT_BLT_X (player_id, cnt)
            y   = CARD_PSTN_PLYR_PNT_BLT_Y (player_id)

            pnt.append_belt_cards(card)

        elif card.type == OG_CARD_TYPE_COVER:

            cnt = pnt.get_cover_cards_cnt()
            x   = CARD_PSTN_PLYR_PNT_CVR_X (player_id, cnt)
            y   = CARD_PSTN_PLYR_PNT_CVR_Y (player_id, cnt)

            pnt.append_cover_cards(card)

        elif card.type == OG_CARD_TYPE_JOKER:
            # TODO : sohalee 조커등을 구현
            pass

        self.move_card(card, x, y, True)


    # 게임 시작시 화투장을 바닥에 깔기 위해 뒤집는다.
    def open_center_card_for_spread(self):

        card_id   = self.get_curr_card_id_center()
        card      = self.cards_full[card_id]
        infos     = self.get_spread_infos(card.number)
        spread_id = -1

        if infos.has_key(KEY_SPRD_EMPT_ID):
            spread_id = infos[KEY_SPRD_EMPT_ID]

        if infos[KEY_SPRD_MTCH_CNT] > 1:

            spread_id = infos[KEY_SPRD_FRST_ID]
            self.merge_spreads(infos[KEY_SPRD_SCND_ID], spread_id)

        else:
            if infos.has_key(KEY_SPRD_MLTI_ID):
                spread_id = infos[KEY_SPRD_MLTI_ID]

        self.append_spread_card(spread_id, card)

        empty_card_id = self.get_curr_empty_card_id_center()
        self.cards_empty[empty_card_id].hide()
        self.card_ids_center.pop()


    def open_center_card_for_player(self, player_id):

        card_id = self.get_curr_card_id_center()
        empty_card_id = self.get_curr_empty_card_id_center()
        card = self.cards_full[card_id]
        card_empty = self.cards_empty[empty_card_id]

        if player_id == MY_PLYR_ID:
            self.card_click_on(card)
            card_empty.hide()

        else:
            card = card_empty

        cnt = len(self.player_hidden_card_ids[player_id])
        x   = CARD_PSTN_PLYR_HIDN_X(player_id, cnt)
        y   = CARD_PSTN_PLYR_HIDN_Y(player_id, cnt / CARD_COUNT_PLYR_BY_LINE)

        self.move_card(card, x, y, False)

        self.player_hidden_card_ids[player_id].append(card_id)
        self.player_hidden_empty_card_ids[player_id].append(empty_card_id)
        self.card_ids_center.pop()


    def get_card_id_by_card(self, card):

        for i in range(CARD_CNT):

            if self.cards_full[i] == card:
                return i

        return -1


    def get_curr_card_id_center(self):

        if not self.network_play:
            card_id = self.card_ids_center[len(self.card_ids_center) - 1]
            self.state_bean.set_card_id_center(card_id)

        return self.state_bean.get_card_id_center()


    def get_curr_empty_card_id_center(self):

        return len(self.card_ids_center) - 1


    def get_curr_player_id(self):

        return self.state_bean.get_player_id()


    def next_player(self):

        self.state_bean.set_player_id((self.get_curr_player_id() + 1)
                                      % PLYR_CNT)

        return self.get_curr_player_id()


    def data_init(self):

        self.state_bean = OgPlayState()

        # 섞어놓은 중앙패의 인덱스. 전체 화투장의 인덱스이다.
        self.card_ids_center = []

        # 바닥패 공간 비움
        for spread in self.spread_spaces:
            spread.clear()

        # player 패 인덱스
        self.player_hidden_card_ids = []
        self.player_hidden_empty_card_ids = []

        for i in range(PLYR_CNT):
            self.player_hidden_card_ids.append([])

        for i in range(PLYR_CNT):
            self.player_hidden_empty_card_ids.append([])

        # player 점수
        self.player_points = []

        for i in range(PLYR_CNT):
            self.player_points.append(OgPlayPoint())


    # 전체 화투장을 섞는다.
    def mix_card(self):

        new_ids = range(CARD_CNT)

        while new_ids:
            rand_id = random.randint(0, len(new_ids) - 1)
            self.card_ids_center.append(new_ids.pop(rand_id))


    # 카드 전체를 사선으로 쌓아 가운데에 놓는다.
    def remain_init(self):

        for i in range(CARD_CNT):

            x = CARD_PSTN_REMN_X + (i / 4)
            y = CARD_PSTN_REMN_Y + (i / 4)

            card = self.cards_full[i]
            card_empty = self.cards_empty[i]

            card.hide()
            card_empty.show()

            self.fixed.put(card, 0, 0)
            self.fixed.put(card_empty, x, y)


    def distribute_spreads(self):

        for i in range(4):

            gtk.gdk.threads_enter()
            self.open_center_card_for_spread()
            gtk.gdk.threads_leave()
            time.sleep(DST_STEP_TIME)

    def distribute_players(self):

        player_id = self.next_player()

        for i in range(5):
            gtk.gdk.threads_enter()
            self.open_center_card_for_player(player_id)
            gtk.gdk.threads_leave()
            time.sleep(DST_STEP_TIME)

        player_id = self.next_player()

        for i in range(5):
            gtk.gdk.threads_enter()
            self.open_center_card_for_player(player_id)
            gtk.gdk.threads_leave()
            time.sleep(DST_STEP_TIME)

    # 라운드 시작시 화투장을 배분한다.
    def distribute_timeout(self):

        self.distribute_spreads()
        self.distribute_players()
        self.distribute_spreads()
        self.distribute_players()

        # FIXEME: 아래는 테스트하기위해 강제로 나를 선으로 만든것이다 지우자
        player_id = self.next_player()

        self.state_bean.set_submit_lock(False)


    def distribute_init(self):

        self.thread_action(self.distribute_timeout, 0)


    # 라운드 시작을 위해 화투장을 섞고 배분하는 등의 준비 작업을 한다.
    def round_init(self):

        self.data_init()

        if not self.network_play:
            self.mix_card()

        self.remain_init()
        self.distribute_init()
