# -*- coding: utf-8 -*-
# 이미지 버튼
#
#   This file is part of otak-godori.
#
#   Copyright (C) 2007 Otak-godori Project.
#
#   This program is free software; you can redistribute it and/or modify it
#   under the terms of the GNU General Public License as published by the
#   Free Software Foundation; either version 3, or (at your option) any later
#   version.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#   Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  

import gtk


IMAGE_BUTTON_TYPE_COUNT = 4

(
    IMAGE_BUTTON_TYPE_NORMAL,
    IMAGE_BUTTON_TYPE_OVER,
    IMAGE_BUTTON_TYPE_PRESS,
    IMAGE_BUTTON_TYPE_DISABLE
) = range(IMAGE_BUTTON_TYPE_COUNT)


class ImageButton(gtk.DrawingArea):

    def __init__(self, filepath, width, height):

        gtk.DrawingArea.__init__(self)

        self.type    = IMAGE_BUTTON_TYPE_NORMAL
        self.width   = width
        self.height  = height
        self.disable = False

        self.set_size_request(width, height)

        self.connect("expose_event", self.on_expose_event, None)
        self.connect("enter_notify_event", self.on_enter_notify_event, None)
        self.connect("leave_notify_event", self.on_leave_notify_event, None)

        self.set_events(gtk.gdk.ENTER_NOTIFY_MASK
                        | gtk.gdk.LEAVE_NOTIFY_MASK
                        | gtk.gdk.BUTTON_PRESS_MASK
                        | gtk.gdk.BUTTON_RELEASE_MASK)

        self.construct_pixbufs(filepath)


    def construct_pixbufs(self, filepath):

        pixbuf = gtk.gdk.pixbuf_new_from_file(filepath)
        pixbuf_width = pixbuf.get_width()
        pixbuf_height = pixbuf.get_height()

        full_width = self.width * IMAGE_BUTTON_TYPE_COUNT

        if pixbuf_width != full_width or pixbuf_height != self.height:
            raise AssertionError("mismatch image size")


        self.pixbufs = []

        for i in range(IMAGE_BUTTON_TYPE_COUNT):

            x = i * self.width
            pixbuf_el = pixbuf.subpixbuf(x, 0, self.width, self.height)

            self.pixbufs.append(pixbuf_el)


    def on_enter_notify_event(self, widget, event, user_data):

        self.change_type(IMAGE_BUTTON_TYPE_OVER)

        return False


    def on_leave_notify_event(self, widget, event, user_data):

        self.change_type(IMAGE_BUTTON_TYPE_NORMAL)

        return False


    def on_expose_event(self, widget, event, user_data):

        gc = widget.style.fg_gc[widget.state]
        pixbuf = self.pixbufs[self.type]
        widget.window.draw_pixbuf (gc, pixbuf, 0, 0, 0, 0,
                                   self.width, self.height,
                                   gtk.gdk.RGB_DITHER_NORMAL, 0, 0);

        return False


    def change_type(self, type):

        old_type = self.type

        if self.disable:
            self.type = IMAGE_BUTTON_TYPE_DISABLE
        else:
            self.type = type

        if self.type != old_type:
            self.window.invalidate_rect(None, False)
