# -*- coding: utf-8 -*-
# 화투장 위젯 소스
#
#   This file is part of otak-godori.
#
#   Copyright (C) 2007 Otak-godori Project.
#
#   This program is free software; you can redistribute it and/or modify it
#   under the terms of the GNU General Public License as published by the
#   Free Software Foundation; either version 3, or (at your option) any later
#   version.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
#   Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  

import gobject
import gtk

(
    OG_CARD_TYPE_COVER, # 피(껍데기)
    OG_CARD_TYPE_BELT, # 띠
    OG_CARD_TYPE_LIGHT, # 광
    OG_CARD_TYPE_TEN, # 열끗
    OG_CARD_TYPE_JOKER  # 조커
) = range(5) # 기본 타입

(
    OG_CARD_SUB_TYPE_NONE, # 없음. BELT(띠)이고 NONE이면 초단
    OG_CARD_SUB_TYPE_BLUE, # 청단
    OG_CARD_SUB_TYPE_RED, # 홍단
    OG_CARD_SUB_TYPE_BIRD, # 고도리
    OG_CARD_SUB_TYPE_GRASS, # 초단
    OG_CARD_SUB_TYPE_DOUBLE, # 쌍피
    OG_CARD_SUB_TYPE_TRIPLE  # 삼피
) = range(7) # 부 타입


class OgCard(gtk.EventBox):

    def __init__(self):

        gtk.EventBox.__init__(self)

        self.pixbuf = None
        self.number = -1
        self.type = OG_CARD_TYPE_COVER
        self.sub_type = OG_CARD_SUB_TYPE_NONE
        self.mixed = False
        self.fixed = gtk.Fixed()
        self.image_widget = gtk.Image()
        self.highlight_darea = gtk.DrawingArea()
        self.cover_darea = gtk.DrawingArea()
        self.s_handler_id_map = {}
        self.enter_notify_handle = None

        self.fixed.put(self.image_widget, 0, 0)
        self.fixed.put(self.highlight_darea, 0, 0)
        self.fixed.put(self.cover_darea, 0, 0)
        self.fixed.show_all()
        self.add(self.fixed)

        self.highlight_off()
        self.cover_hide()

        self.highlight_darea.connect("expose_event", 
                                     self.on_highlight_expose_event, None)
        self.cover_darea.connect("expose_event", 
                                 self.on_cover_expose_event, None)


    # GObject connect override
    def connect(self, detailed_signal, handler, user_data):

        self.disconnect_by_func(handler)

        handler_id = gobject.GObject.connect(self, detailed_signal, handler, 
                                             user_data)
        self.s_handler_id_map[handler] = handler_id

        return handler_id


    # 함수로 disconnect를 삭제
    def disconnect_by_func(self, handler):

        if self.s_handler_id_map.has_key(handler):
            handler_id = self.s_handler_id_map[handler]
            self.disconnect(handler_id)
            del self.s_handler_id_map[handler]


    # signal callback. 마우스 포인터 진입시 강조 위젯이 보여지게 한다.
    def on_enter_notify_event(self, widget, event, user_data):

        if self.highlight_cb:
            self.highlight_cb(self)

        self.highlight_darea.show()

        return False


    # signal callback. 강조 위젯 expose callback.
    def on_highlight_expose_event(self, widget, event, user_data):

        width    = widget.allocation.width
        height   = widget.allocation.height
        black    = widget.get_colormap().alloc_color("black")
        white    = widget.get_colormap().alloc_color("white")
        selected = widget.get_colormap().alloc_color("blue")
        mask     = gtk.gdk.Pixmap(widget.window, width, height, 1)
        maskgc   = mask.new_gc()
        gc       = widget.window.new_gc()

        gc.set_foreground(selected)
        widget.window.draw_rectangle (gc, True, 0, 0, width, height)
        maskgc.set_foreground(white)
        mask.draw_rectangle (maskgc, True, 0, 0, width, height)
        maskgc.set_foreground(black)
        mask.draw_rectangle (maskgc, True, 3, 3, width - 6, height - 6)

        widget.shape_combine_mask (mask, 0, 0);

        return False


    # signal callback. 덮개 위젯 expose callback.
    # 화투장 가리개로도 사용한다.
    def on_cover_expose_event(self, widget, event, user_data):

        if self.cover_pixbuf:
            width  = widget.allocation.width
            height = widget.allocation.height
            gc     = widget.style.fg_gc[widget.state]

            widget.window.draw_pixbuf(gc, self.cover_pixbuf, 0, 0, 0, 0,
                width, height, gtk.gdk.RGB_DITHER_NORMAL, 0, 0)

        return False


    # 화투패 이미지 셋팅
    def set_from_file(self, filepath):

        self.pixbuf = gtk.gdk.pixbuf_new_from_file(filepath)
        self.image_widget.set_from_pixbuf(self.pixbuf)


    # 번호. 1~12사이
    def set_number(self, number):

        self.number = number


    # 형식 속성 정의. (광, 피 등)
    def set_type(self, type, sub_type):

        self.type     = type
        self.sub_type = sub_type


    # 덮개 표시
    def cover_show(self):

        darea = self.cover_darea
        darea.show()

        if darea.flags() & gtk.REALIZED:
            darea.window.invalidate_rect(None, False)


    # 덮개 숨김
    def cover_hide(self):

        self.cover_darea.hide()


    # 덮개에 이미지 셋팅
    def set_cover_pixbuf(self, pixbuf):

        self.cover_pixbuf = pixbuf


    # 넓이와 높이를 지정한다.
    def set_size_request(self, width, height):

        gtk.Widget.set_size_request(self, width, height)
        gtk.Widget.set_size_request(self.image_widget, width, height)
        gtk.Widget.set_size_request(self.highlight_darea, width, height)
        gtk.Widget.set_size_request(self.cover_darea, width, height)
        gtk.Widget.set_size_request(self.fixed, width, height)

        if self.pixbuf:
            pixbuf = self.pixbuf.scale_simple(width, height, 
                                              gtk.gdk.INTERP_BILINEAR)
            self.image_widget.set_from_pixbuf(pixbuf)


    # 강조 위젯에 마우스 포인터 진입시 반응하도록 signal을 연결한다.
    def highlight_connect(self, highlight_cb):

        self.enter_notify_handle = self.connect("enter_notify_event", 
                                                self.on_enter_notify_event,
                                                None)
        self.highlight_cb = highlight_cb


    # 강조 위젯의 마우스 포인터 진입시 반응하는 signal 제거
    def highlight_disconnect(self):

        if self.enter_notify_handle:
            self.disconnect(self.enter_notify_handle)
            self.enter_notify_handle = None


    # 강조 기능 끄기
    def highlight_off(self):

        self.highlight_darea.hide()
