/* 게임 진행 소스

   This file is part of otak-godori.

   Copyright (C) 2007 Otak-godori Project.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  
*/

#include <stdlib.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <glade/glade-xml.h>

#include "widget/ogcard.h"
#include "widget/ogbutton.h"
#include "otak-lib/otak-util/otsys.h"
#include "ogdefine.h"
#include "ogconf.h"
#include "ogplayui.h"

#define CARD_COUNT              48  /* 총 화투장 수 */
#define SPRD_COUNT              16  /* 바닥에 화투장 깔 공간 수 */
#define SPRD_CARD_COUNT         6   /* 바닥에 화투장 깔 공간당 화투장 수 */
#define PLAYER_COUNT            2   /* 플레이어 수 */
#define PLAYER_CARD_COUNT       10  /* 플레이어당 가지는 숨겨진 화투장 수 */
#define PLAYER_PNT_LIGHT_COUNT  5   /* 점수패 광 최대 수 */
#define PLAYER_PNT_TEN_COUNT    9   /* 점수패 열끗 최대수 */
#define PLAYER_PNT_BELT_COUNT   10  /* 점수패 띠 최대수 */
#define PLAYER_PNT_COVER_COUNT  27  /* 점수패 피(껍데기) 최대수 */
#define DST_STEP_TIME           300 /* 라운드 시작시 화투장 나누어줄때 딜레이 시간. 1/1000 second */
#define MY_PLAYER_ID            1   /* 내 플레이어 인덱스. 선, 후를 따니는게 아니고 위아래 공간 정도 */

#define CARD_RAIN_NUMBER        12  /* 화투장 비의 카드번호 */

#define VALID_POINT             7   /* 유효 점수 (나는 점수) */

/* 화투장 나누어주는 진행된 단계 */
typedef enum
{
    DST_STEP_NONE,
    DST_STEP_SPRD_1,    /* 바닥 첫번째 나눠주기 */
    DST_STEP_SPRD_2,    /* 바닥 두번째 나눠주기 */
    DST_STEP_PLYR_1_1,  /* 플레이어 1 첫번째 나눠주기 */
    DST_STEP_PLYR_1_2,  /* 플레이어 1 두번째 나눠주기 */
    DST_STEP_PLYR_2_1,  /* 플레이어 2 첫번째 나눠주기 */
    DST_STEP_PLYR_2_2   /* 프렐이어 2 두번째 나눠주기 */
} DstStep;

typedef enum
{
    SUBMIT_TYPE_NONE,
    SUBMIT_TYPE_PLAYER,
    SUBMIT_TYPE_OPEN
} SubmitType;

/* 화투장 타입 */
typedef struct {
  OgCardType    type;
  OgCardSubType sub_type;
} CardType;

/*
 * 화투장 배열에 각각 타입 정의
 */
static CardType CARD_TYPES[CARD_COUNT] = {
    /* 1 송학 */
    { OG_CARD_TYPE_LIGHT,   OG_CARD_SUB_TYPE_NONE },    // 광
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_RED },     // 띠, 청단
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 2 매조 */
    { OG_CARD_TYPE_TEN,     OG_CARD_SUB_TYPE_BIRD },    // 열끗, 고도리
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_RED },     // 띠, 홍단
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 3 벚꽃 */
    { OG_CARD_TYPE_LIGHT,   OG_CARD_SUB_TYPE_NONE },    // 광
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_RED },     // 띠, 홍단
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 4 흑싸리 */
    { OG_CARD_TYPE_TEN,     OG_CARD_SUB_TYPE_BIRD },    // 열끗, 고도리 
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_GRASS },   // 띠, 초단
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 5 난초 */
    { OG_CARD_TYPE_TEN,     OG_CARD_SUB_TYPE_NONE },    // 열끗
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_GRASS },   // 띠, 초단
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 6 모란 */
    { OG_CARD_TYPE_TEN,     OG_CARD_SUB_TYPE_NONE },    // 열끗
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_BLUE },    // 띠, 청단
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 7 홍싸리 */
    { OG_CARD_TYPE_TEN,     OG_CARD_SUB_TYPE_NONE },    // 열끗
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_GRASS },   // 띠, 초단
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 8 공산 */
    { OG_CARD_TYPE_LIGHT,   OG_CARD_SUB_TYPE_NONE },    // 광
    { OG_CARD_TYPE_TEN,     OG_CARD_SUB_TYPE_BIRD },    // 열끗, 고도리
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 9 국준 */
    { OG_CARD_TYPE_TEN,     OG_CARD_SUB_TYPE_DOUBLE },  // 열끗, 쌍피
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_BLUE },    // 띠, 청단
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 10 단풍 */
    { OG_CARD_TYPE_TEN,     OG_CARD_SUB_TYPE_NONE },    // 열끗
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_BLUE },    // 띠, 청단
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 11 오동 */
    { OG_CARD_TYPE_LIGHT,   OG_CARD_SUB_TYPE_NONE },    // 광
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_DOUBLE },  // 피, 쌍피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_NONE },    // 피
    /* 12 비 */
    { OG_CARD_TYPE_LIGHT,   OG_CARD_SUB_TYPE_NONE },    // 광(비광)
    { OG_CARD_TYPE_TEN,     OG_CARD_SUB_TYPE_NONE },    // 열끗, 
    { OG_CARD_TYPE_BELT,    OG_CARD_SUB_TYPE_NONE },    // 띠(초단아님)
    { OG_CARD_TYPE_COVER,   OG_CARD_SUB_TYPE_DOUBLE }   // 피, 쌍피
    
//    { OG_CARD_TYPE_JOKER,   OG_CARD_SUB_TYPE_DOUBLE },  // 조커, 투피
//    { OG_CARD_TYPE_JOKER,   OG_CARD_SUB_TYPE_TRIPLE }   // 조커, 삼피
};

/* 화투장 공간 위치 (상대적인 위치) */
typedef struct {
  gint  x;
  gint  y;
} CardPosition;

/* 화투 바닥패 위치 */
static CardPosition SPREAD_POSITION[SPRD_COUNT] = {
    /* top */
    { CARD_PSTN_OPEN_TOP_X(0),      CARD_PSTN_OPEN_TOP_Y },
    { CARD_PSTN_OPEN_TOP_X(1),      CARD_PSTN_OPEN_TOP_Y },
    /* right */
    { CARD_PSTN_OPEN_RIGHT_X,       CARD_PSTN_OPEN_RIGHT_Y(0) },
    { CARD_PSTN_OPEN_RIGHT_X,       CARD_PSTN_OPEN_RIGHT_Y(1) },
    /* bottom */
    { CARD_PSTN_OPEN_BOTTOM_X(1),   CARD_PSTN_OPEN_BOTTOM_Y },
    { CARD_PSTN_OPEN_BOTTOM_X(0),   CARD_PSTN_OPEN_BOTTOM_Y },
    /* left */
    { CARD_PSTN_OPEN_LEFT_X,        CARD_PSTN_OPEN_LEFT_Y(1) },
    { CARD_PSTN_OPEN_LEFT_X,        CARD_PSTN_OPEN_LEFT_Y(0) },
    /* 8장을 초과 top */
    { CARD_PSTN_OPEN_LEFT_X,        CARD_PSTN_OPEN_LEFT_Y(0)
                                        - CARD_NML_HEIGHT
                                        - CARD_SPC_OPEN },
    { CARD_PSTN_OPEN_RIGHT_X,       CARD_PSTN_OPEN_RIGHT_Y(0)
                                        - CARD_NML_HEIGHT
                                        - CARD_SPC_OPEN },
    /* 8장 초과 right */
    { CARD_PSTN_OPEN_RIGHT_X
        + CARD_NML_WIDTH
        + CARD_SPC_OPEN,            CARD_PSTN_OPEN_RIGHT_Y(0) },
    { CARD_PSTN_OPEN_RIGHT_X
        + CARD_NML_WIDTH
        + CARD_SPC_OPEN,            CARD_PSTN_OPEN_RIGHT_Y(1) },
    /* 8장 초과 bottom */
    { CARD_PSTN_OPEN_RIGHT_X,       CARD_PSTN_OPEN_RIGHT_Y(2) },
    { CARD_PSTN_OPEN_LEFT_X,        CARD_PSTN_OPEN_LEFT_Y(2) },
    /* 8장 초과 left */
    { CARD_PSTN_OPEN_LEFT_X
        - CARD_NML_WIDTH
        - CARD_SPC_OPEN,            CARD_PSTN_OPEN_LEFT_Y(1) },
    { CARD_PSTN_OPEN_LEFT_X
        - CARD_NML_WIDTH
        - CARD_SPC_OPEN,            CARD_PSTN_OPEN_LEFT_Y(0) }
};

typedef struct {
    gint    card_id[SPRD_CARD_COUNT];
    gint    count;
    gint    joker_count;
} CardSpread;

typedef struct {
    gint    empty_card_ids[PLAYER_CARD_COUNT];
    gint    hidden_card_ids[PLAYER_CARD_COUNT];
    gint    submit_cnt;
} CardPlayer;

typedef struct {
    gint    light_card_ids[PLAYER_PNT_LIGHT_COUNT];
    gint    ten_card_ids[PLAYER_PNT_TEN_COUNT];
    gint    belt_card_ids[PLAYER_PNT_BELT_COUNT];
    gint    cover_card_ids[PLAYER_PNT_COVER_COUNT];
    gint    light_cnt;
    gint    ten_cnt;
    gint    belt_cnt;
    gint    cover_cnt;

    gint    bird_cnt;
    gint    blue_cnt;
    gint    red_cnt;
    gint    grass_cnt;
    gint    double_cnt;
    gint    triple_cnt;

    gint    go_step;
    gint    go_point;
} CardPoint;

static gboolean     init                = FALSE;
static gboolean     network_play        = FALSE;

static GladeXML     *play_glade_xml     = NULL;
static OgCard       **og_cards_empty    = NULL; /* 빈패 (상대패나 뒤집힌 패등 안보이는 패를 표시) */
static OgCard       **og_cards          = NULL; /* 화투장이 모두 들어있는 기본 자료 */
static OgCard       *og_card_highlight  = NULL; /* 현재 highlight(내 화투장에 마우스 오버하면 테두리생기는거) */

static gint         card_all_mixed_ids[CARD_COUNT];         /* 화투장 전부 섞여놓은 인덱스 */

static CardSpread   spreads[SPRD_COUNT];                    /* 바닥 화투장 공간 */

static guint        dst_timeout_handler = 0;                /* 화투장을 멋지게 나누어주기위해 타이머 이용 */
static DstStep      dst_step            = DST_STEP_NONE;    /* 현재까지 나누어준 화투장 단계 */

static CardPlayer   player_cards[PLAYER_COUNT];             /* 플레이어의 화투장 정보. 가진패, 먹은패 모두 포함 */
static CardPoint    player_point[PLAYER_COUNT];
static gboolean     player_submit_lock          = TRUE;     /* 플레이어 패 제출 잠금. TRUE면 화투장을 못냄 */
static gint         player_first_id             = 0;        /* 플레이어중 선의 index */

static gint         curr_open_card_cnt  = 0;                /* 현재까지 열려진 카드의 수. 바닥패와 플레이어패 포함 */
static gint         curr_top_card_id    = CARD_COUNT - 1;   /* 가운데 패 맨위 화투장 전체에서의 인덱스.
                                                               로칼에선 필요없지만 네트웍플레이에서는 mixed등을 사용하지 않고 수시로 값을 받는다. */

static gint         curr_player_id              = -1;   /* 현재 진행중이 플레이어의 아이디 */
static gint         curr_match_spread_ids[2];           /* 현재 패가 매칭되어 완전히 놓여진 공간의 index */

static SubmitType   curr_submit_type                = SUBMIT_TYPE_NONE; /* 현재 패를 낸것이 플레이어인가 뒤집은것인가 */
static gint         curr_submit_match_spread_ids[2];                    /* 플레이어가 내거나 뒤집은 카드와 같은 번호의 바닥패 공간의 index */
static gint         curr_submit_card_id             = -1;               /* 플레이어가 제출하거나 뒤집은 패 */

/*
 * method
 */
static void         og_play_ui_card_highlight_cb                (OgCard *card);

static GladeXML*    og_play_ui_get_glade_xml                    (void);
static GtkWidget*   og_play_ui_get_fixed                        (void);
static GtkWidget*   og_play_ui_get_card_choice_window           (void);
static GtkWidget*   og_play_ui_get_card_choice_image_1          (void);
static GtkWidget*   og_play_ui_get_card_choice_image_2          (void);
static GtkWidget*   og_play_ui_get_go_choice_window             (void);
static GtkWidget*   og_play_ui_get_go_choice_fixed              (void);

static gint         og_play_ui_get_relative_rand_card_id        (gint rand_id);
static void         og_play_ui_next_top_card_by_mixed           (void);
static OgCard*      og_play_ui_get_top_card                     (void);
static gint         og_play_ui_get_empty_top_card_id            (void);
static gint         og_play_ui_get_spreads_using_count          (void);

static gint         og_play_ui_get_card_id_by_og_card           (OgCard *value);
static gint         og_play_ui_next_player_id                   (void);

static gint         og_play_ui_card_player_get_curr_id          (CardPlayer *player);

static gboolean     og_play_ui_player_is_submit_lock            (void);
static void         og_play_ui_player_submit_lock               (void);
static void         og_play_ui_player_submit_unlock             (void);
static void         og_play_ui_card_click_on                    (OgCard *card);
static void         og_play_ui_card_click_off                   (OgCard *card);

static void         og_play_ui_card_spread_clear                (CardSpread *spread);
static void         og_play_ui_card_spreads_clear               (CardSpread spreads[],
                                                                 gint       count);
static void         og_play_ui_card_player_clear                (CardPlayer *player);
static void         og_play_ui_curr_submit_clear                (void);
static void         og_play_ui_curr_match_spread_ids_clear      (void);

static gint         og_play_ui_card_point_get_total_point       (CardPoint *point);
static gint         og_play_ui_card_point_get_total_point_final (CardPoint *point);

static void         og_play_ui_card_move                        (OgCard     *card,
                                                                 gint        x,
                                                                 gint        y,
                                                                 gboolean    resent);

static void         og_play_ui_spread_infos_by_card_number      (gint    card_number,
                                                                 gint   *result_count,
                                                                 gint   *result_id_from_first_empty,
                                                                 gint   *result_spread_id_1,
                                                                 gint   *result_spread_id_2,
                                                                 gint   *result_single_spread_id,
                                                                 gint   *result_multi_spread_id);
static void         og_play_ui_spread_add_card                  (gint   spread_id,
                                                                 gint   card_id);
static void         og_play_ui_spread_merge_by_card_number      (gint   dest_spread_id,
                                                                 gint   src_spread_id);

static void         og_play_ui_card_choice_show                 (void);
static void         og_play_ui_card_choice_done                 (gint choid_id);

static void         og_play_ui_submit                           (gint card_id);
static void         og_play_ui_submit_done                      (gint   spread_id,
                                                                 gint   card_id);
static void         og_play_ui_submit_pickup                    (gint player_id);

static void         og_play_ui_player_submit_card               (gint player_id,
                                                                 gint card_id);
static void         og_play_ui_player_add_hidden_card           (gint player_id);
static void         og_play_ui_player_add_point_card            (gint   player_id,
                                                                 gint   card_id);

static void         og_play_ui_open_card_by_player              (void);
static void         og_play_ui_open_card_for_spread             (void);


/*
 * signal
 */

/* play window delete */
gboolean
og_play_window_delete_event_cb (GtkWidget  *widget,
                                GdkEvent   *event,
                                gpointer    user_data)
{
    gboolean noquit = FALSE;
    GtkWidget *dialog = gtk_dialog_new_with_buttons ("Otak Godori Quit?",
                                                     GTK_WINDOW (og_play_ui_get_window ()),
                                                     GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                                                     GTK_STOCK_OK,
                                                     GTK_RESPONSE_OK,
                                                     GTK_STOCK_CANCEL,
                                                     GTK_RESPONSE_CANCEL,
                                                     NULL);

    gint result = gtk_dialog_run (GTK_DIALOG (dialog));
    switch (result)
    {
        case GTK_RESPONSE_OK:
            gtk_main_quit ();
            break;

        default:
            noquit = TRUE;
            break;
    }

    gtk_widget_destroy (dialog);
    dialog = NULL;

    return noquit;
}

/* play fixed enter notify */
gboolean
og_play_eventbox_enter_notify_event_cb (GtkWidget          *widget,
                                        GdkEventCrossing   *event,
                                        gpointer            user_data)
{
    og_play_ui_card_highlight_cb (NULL);
    
    return FALSE;
}

gboolean
og_play_ui_player_card_hidden_button_press_event_cb (GtkWidget        *widget,
                                                     GdkEventButton   *event,
                                                     gpointer          user_data)
{
    if (!og_play_ui_player_is_submit_lock ())
    {
        gint card_id = og_play_ui_get_card_id_by_og_card (OG_CARD (widget)); 

        og_play_ui_player_submit_card (curr_player_id, card_id);
    }

    return FALSE;
}

/*
 * 중복 패중 1 선택 윈도우 관련 signal
 */
gboolean
og_play_card_choice_eventbox_1_button_press_event_cb (GtkWidget         *widget,
                                                      GdkEventButton    *event,
                                                      gpointer           user_data)
{
    og_play_ui_card_choice_done (0);

    return FALSE;
}

gboolean
og_play_card_choice_eventbox_2_button_press_event_cb (GtkWidget         *widget,
                                                      GdkEventButton    *event,
                                                      gpointer           user_data)
{
    og_play_ui_card_choice_done (1);

    return FALSE;
}

/*
 * custom signal
 */

/* OgCard가 highlight 표시를 한 후에 실행할 Callback */
static void
og_play_ui_card_highlight_cb (OgCard *card)
{
    if (og_card_highlight != NULL)
    {
        og_card_highlight_off (og_card_highlight);
    }

    og_card_highlight = card;
}



/*
 * private
 */

/*
 * og_play_ui_get_glade_xml:
 * 
 * 게임진행 화면의 위젯을 호출할 수 있는 GladeXML* 을 반환한다.
 * 
 * Return value: libglade용 GladeXML* 또는 %NULL
 */
static GladeXML*
og_play_ui_get_glade_xml (void)
{
    if (play_glade_xml == NULL)
    {
        play_glade_xml = glade_xml_new ("glade/og_play.glade", NULL, NULL);
        glade_xml_signal_autoconnect (play_glade_xml);
    }

    return play_glade_xml;
}

/*
 * og_play_ui_get_fixed:
 * 
 * 게임진행 화면의 배치용 Container로 사용하는 Fixed를 반환한다.
 * 
 * Return value: 바탕격인 위젯 또는 %NULL
 */
static GtkWidget*
og_play_ui_get_fixed (void)
{
    return glade_xml_get_widget (og_play_ui_get_glade_xml(), "og_play_fixed");
}

/*
 * og_play_ui_get_card_choice_window:
 * 
 * 바닥패중 플레이어가 낸 패랑 같은 번호가 두개 있을때 둘중 하나를 선택하는 Window 위젯을 반환
 * 
 * Return value: 중복 바닥패 선택 Window 위젯
 */
static GtkWidget*
og_play_ui_get_card_choice_window (void)
{
    return glade_xml_get_widget (og_play_ui_get_glade_xml(), "og_play_card_choice_window");
}

/*
 * og_play_ui_get_card_choice_image_1:
 * 
 * 중복 바닥패 선택 Window의 첫번째 화투장 image 위젯을 반환
 * 
 * Return value: 첫번째 화투장 image 위젯
 */
static GtkWidget*
og_play_ui_get_card_choice_image_1 (void)
{
    return glade_xml_get_widget (og_play_ui_get_glade_xml(), "og_play_card_choice_image_1");
}

/*
 * og_play_ui_get_card_choice_image_2:
 * 
 * 중복 바닥패 선택 Window의 두번째 화투장 image 위젯을 반환
 * 
 * Return value: 두번째 화투장 image 위젯
 */
static GtkWidget*
og_play_ui_get_card_choice_image_2 (void)
{
    return glade_xml_get_widget (og_play_ui_get_glade_xml(), "og_play_card_choice_image_2");
}

/*
 * og_play_ui_get_go_choice_window:
 * 
 * 고와 스톱을 선택하는 Window 위젯을 반환
 * 
 * Return value: 고, 스톱 선택 Window 위젯
 */
static GtkWidget*
og_play_ui_get_go_choice_window (void)
{
    return glade_xml_get_widget (og_play_ui_get_glade_xml(), "og_play_go_choice_window");
}

/*
 * og_play_ui_get_go_choice_fixed:
 * 
 * 고와 스톱을 선택하는 창의 Fixed 위젯을 반환
 * 
 * Return value: 고, 스톱 선택 Fixed 위젯
 */
static GtkWidget*
og_play_ui_get_go_choice_fixed (void)
{
    return glade_xml_get_widget (og_play_ui_get_glade_xml(), "og_play_go_choice_fixed");
}

/*
 * og_play_ui_get_relative_rand_card_id:
 * @rand_id: 섞이기 전인 화투장들만을 기준으로한 상대적 index
 * 
 * 화투장 전체를 섞을때 전체 화투장중 아직 섞이지 않는 화투장의 index를 반환한다.
 * rand_id는 현재 섞이지 않은 화투장중에서의 상대적인 index이다.
 * 전체 화투장에서의 인덱스 1, 3, 5가 남은상태에서 rand_id가 1이면 3, 2이면 5
 * 
 * Return value: 전체 화투장 중에서의 index
 */
static gint
og_play_ui_get_relative_rand_card_id (gint rand_id)
{
    gint curr_id = -1;

    gint i = 0;
    for (i = 0; i < CARD_COUNT; i++)
    {
        if (og_cards[i]->mixed == FALSE)
        {
            curr_id++;
        }

        if (curr_id == rand_id)
        {
            og_cards[i]->mixed = TRUE;
            return i;
        }
    }
    
    return -1;
}

static void
og_play_ui_next_top_card_by_mixed (void)
{
    curr_open_card_cnt++;
    if (!network_play)
    {
        curr_top_card_id = card_all_mixed_ids[CARD_COUNT - 1 - curr_open_card_cnt];
    }
}

/*
 * og_play_ui_get_top_card:
 * 
 * 가운데 뒤집어져있는 화투장중 가장 위에 놓여진 화투장의 포인터
 * 
 * Return value: 화투장의 포인터
 */
static OgCard*
og_play_ui_get_top_card (void)
{
    return og_cards[curr_top_card_id];
}

static gint
og_play_ui_get_empty_top_card_id (void)
{
    return CARD_COUNT - 1 - curr_open_card_cnt;
}

/*
 * og_play_ui_get_spreads_using_count:
 * 
 * 바닥 공간중 현재 화투장이 놓여져 있는 사용중인 공간의 수
 * 
 * Return value: 사용중인 바닥공간
 */
static gint
og_play_ui_get_spreads_using_count (void)
{
    gint count = 0;
    
    gint i = 0;
    for (i = 0; i < SPRD_COUNT; i++)
    {
        if (spreads[i].count > 0)
        {
            count++;
        }
    }
    
    return count;
}

/*
 * og_play_ui_player_get_hidden_id_by_og_cards:
 * @value: 화투장의 포인터
 * 
 * 전체 화투패중 value에 해당하는 index를 반환한다.
 * 
 * Return value: 전체 화투패에서의 index
 */
static gint
og_play_ui_get_card_id_by_og_card (OgCard *value)
{
    gint i = 0;
    for (i = 0; i < CARD_COUNT; i++)
    {
        if (og_cards[i] == value)
        {
            return i;
        }
    }

    return -1;
}

static gint
og_play_ui_next_player_id (void)
{
    curr_player_id = (curr_player_id + 1) % PLAYER_COUNT;
    return curr_player_id;
}

static gint
og_play_ui_card_player_get_curr_id (CardPlayer *player)
{
    return PLAYER_CARD_COUNT - 1 - player->submit_cnt;
}

/*
 * og_play_ui_player_is_submit_lock:
 * 
 * 플레이어가 패를 낼수 없는 상태인지 확인한다.
 * 
 * Return value: TRUE 패를 낼수 없음, FALSE 패를 낼수 있음
 */
static gboolean
og_play_ui_player_is_submit_lock (void)
{
    return player_submit_lock;
}

/*
 * og_play_ui_player_submit_lock:
 * 
 * 플레이어가 패를 낼수 없도록 락을 건다.
 */
static void
og_play_ui_player_submit_lock (void)
{
    player_submit_lock = TRUE;
}

/*
 * og_play_ui_player_submit_unlock:
 * 
 * 플레이어가 패를 낼수 있도록 락을 푼다.
 */
static void
og_play_ui_player_submit_unlock (void)
{
    player_submit_lock = FALSE;
}

/*
 * og_play_ui_card_click_on:
 * @card: 클릭할수 있도록 만들 화투장
 * 
 * 특정화투를 클릭할수 있는 상태로 만든다. (자기패)
 */
static void
og_play_ui_card_click_on (OgCard *card)
{
    og_card_highlight_connect (card, og_play_ui_card_highlight_cb);
    g_signal_connect ((gpointer) card, "button_press_event",
            G_CALLBACK (og_play_ui_player_card_hidden_button_press_event_cb), NULL);
}

/*
 * og_play_ui_card_click_off:
 * @card: 클릭할 수 없도록 만들 화투장
 * 
 * 특정화투를 클릭할수 없는 상태로 만든다. (자기패가 아닌것)
 */
static void
og_play_ui_card_click_off (OgCard *card)
{
    og_card_highlight_disconnect (card);
    g_signal_handlers_disconnect_by_func ((gpointer) card,
            G_CALLBACK (og_play_ui_player_card_hidden_button_press_event_cb), NULL);
    og_card_highlight_off (card);
}

/*
 * og_play_ui_card_spread_clear:
 * @spread: 초기화 할 바닥패 공간
 * 
 * 화투장 그룹 구조체 멤버의 값 초기화
 */
static void
og_play_ui_card_spread_clear (CardSpread *spread)
{
    gint i = 0;

    /* 열린 화투장 (바닥패) 그룹 초기화 */
    for (i = 0; i < SPRD_CARD_COUNT; i++)
    {
        spread->card_id[i] = -1;
    }

    spread->count = 0;
    spread->joker_count = 0;
}

/*
 * og_play_ui_card_spread_clear:
 * @group[]: 초기화 할 화투장 그룹 구조체들
 * @count: 그룹 구조체의 수
 * 
 * 화투장 그룹 구조체들 멤버의 값 초기화
 */
static void
og_play_ui_card_spreads_clear (CardSpread   spreads[],
                              gint          count)
{
    gint i = 0;

    /* 열린 화투장 (바닥패) 그룹 초기화 */
    for (i = 0; i < count; i++)
    {
        og_play_ui_card_spread_clear (&spreads[i]);
    }
}

static void
og_play_ui_card_player_clear (CardPlayer *player)
{
    gint i = 0;

    for (i = 0; i < PLAYER_CARD_COUNT; i++) {
        player->empty_card_ids[i] = -1;
        player->hidden_card_ids[i] = -1;
    }

    player->submit_cnt      = 0;
}

static void
og_play_ui_card_point_clear (CardPoint *point)
{
    gint i = 0;

    for (i = 0; i < PLAYER_PNT_LIGHT_COUNT; i++) {
        point->light_card_ids[i] = -1;
    }

    for (i = 0; i < PLAYER_PNT_TEN_COUNT; i++) {
        point->ten_card_ids[i] = -1;
    }

    for (i = 0; i < PLAYER_PNT_BELT_COUNT; i++) {
        point->belt_card_ids[i] = -1;
    }

    for (i = 0; i < PLAYER_PNT_COVER_COUNT; i++) {
        point->cover_card_ids[i] = -1;
    }

    point->light_cnt = 0;
    point->ten_cnt   = 0;
    point->belt_cnt  = 0;
    point->cover_cnt = 0;

    point->bird_cnt     = 0;
    point->blue_cnt     = 0;
    point->red_cnt      = 0;
    point->grass_cnt    = 0;
    point->double_cnt   = 0;
    point->triple_cnt   = 0;

    point->go_step   = 0;
    point->go_point  = 0;
}

/*
 * og_play_ui_match_spreads_clear:
 * 
 * 글로벌 변수 match_spred_grp_ids를 초기화한다.
 */
static void
og_play_ui_curr_submit_clear (void)
{
    gint i = 0;
    for (i = 0; i < G_N_ELEMENTS (curr_submit_match_spread_ids); i++) {
        curr_submit_match_spread_ids[i] = -1;
    }

    curr_submit_type    = SUBMIT_TYPE_NONE;
    curr_submit_card_id = -1;
}


static void
og_play_ui_curr_match_spread_ids_clear (void)
{
    gint i = 0;
    for (i = 0; i < G_N_ELEMENTS (curr_match_spread_ids); i++) {
        curr_match_spread_ids[i] = -1;
    }
}

static gint
og_play_ui_card_point_get_light_point (CardPoint *point)
{
    gint        return_point    = 0;
    gboolean    rain            = FALSE;

    int i = 0;
    for (i = 0; i < point->light_cnt; i++)
    {
        OgCard *card = og_cards[point->light_card_ids[i]];
        if (card->number == CARD_RAIN_NUMBER)
        {
            rain = TRUE;
        }
    }

    if (point->light_cnt >= 3)
    {
        switch (point->light_cnt)
        {
            case 3:
                return_point = rain ? 2 : 3;
                break;
            case 5:
                return_point = 15;
                break;
            default:
                return_point = point->light_cnt;
                break;
        }
    }

    return return_point;
}

static gint
og_play_ui_card_point_get_ten_point (CardPoint *point)
{
    return point->ten_cnt >= 5 ? point->ten_cnt - 4 : 0;
}

static gint
og_play_ui_card_point_get_belt_point (CardPoint *point)
{
    return point->belt_cnt >= 5 ? point->belt_cnt - 4 : 0;
}

static gint
og_play_ui_card_point_get_cover_point (CardPoint *point)
{
    gint abs_count = point->cover_cnt + point->double_cnt + (point->triple_cnt * 2);

    return abs_count >= 10 ? abs_count - 9 : 0;
}

static gint
og_play_ui_card_point_get_bird_point (CardPoint *point)
{
    return point->bird_cnt == 3 ? 5 : 0;
}

static gint
og_play_ui_card_point_get_blue_point (CardPoint *point)
{
    return point->blue_cnt == 3 ? 3 : 0;
}

static gint
og_play_ui_card_point_get_red_point (CardPoint *point)
{
    return point->red_cnt == 3 ? 3 : 0;
}

static gint
og_play_ui_card_point_get_grass_point (CardPoint *point)
{
    return point->grass_cnt == 3 ? 3 : 0;
}
/*
 * og_play_ui_card_point_get_total_point
 * @point: 합계를 구할 포인트 기초정보
 * 
 * 한 사람의 포인트 전체점수. 배수(멍따, 흔들기, 피박)를 고려하지 않은 순수한 점수임
 * 
 * Return value: 배수를 고려하지 않은 순수한 총 점수
 */
static gint
og_play_ui_card_point_get_total_point (CardPoint *point)
{
    gint return_point = 0;
    
    return_point += og_play_ui_card_point_get_light_point (point);
    return_point += og_play_ui_card_point_get_ten_point (point);
    return_point += og_play_ui_card_point_get_belt_point (point);
    return_point += og_play_ui_card_point_get_cover_point (point);
    return_point += og_play_ui_card_point_get_bird_point (point);
    return_point += og_play_ui_card_point_get_blue_point (point);
    return_point += og_play_ui_card_point_get_red_point (point);
    return_point += og_play_ui_card_point_get_grass_point (point);
    
    return return_point;
}

static gint
og_play_ui_card_point_get_total_point_final (CardPoint *point)
{
    gint return_final = og_play_ui_card_point_get_total_point (point);

    // 멍따
    if (point->ten_cnt >= 7) {
        return_final *= 2;
    }

    return return_final;
}

/*
 * og_play_ui_card_move:
 * @card: 화투장 오브젝트
 * @x: 이동할 x좌표. 게임 화면에서의 상대 좌표
 * @y: 이동할 y좌표. 게임 화면에서의 상대 좌표
 * @resent: 화투장을 다른 화투장 보다 위에 보이게 할지 여부
 * @empty: 화면에 빈패로 보여줄지 여부
 * 
 * 화투장 위젯의 화면 위치를 이동 시킨다. resent가 true인 경우 주변 카드들 보다 위에 보이게끔 한다.
 */
static void
og_play_ui_card_move (OgCard    *card,
                      gint       x,
                      gint       y,
                      gboolean   resent)
{
    GtkWidget *fixed = og_play_ui_get_fixed ();

    if (resent)
    {
        g_object_ref_sink (card);
        gtk_container_remove (GTK_CONTAINER (fixed), GTK_WIDGET (card));
        gtk_fixed_put (GTK_FIXED (fixed), GTK_WIDGET (card), x, y);
        g_object_unref (card);
    }
    else
    {
        gtk_fixed_move (GTK_FIXED (fixed), GTK_WIDGET (card), x, y);
    }

    gtk_widget_show (GTK_WIDGET (card));
}

/*
 * og_play_ui_spread_infos_by_card_number:
 * @card_number: 화투장 번호
 * @result_count: 일치하는 바닥패 공간의 수
 * @result_id_from_first_empty: 바닥패중 비어있는 첫번째 공간의 아이디
 * @spread_id_1: 일치하는 바닥패 공간의 첫번째 아이디
 * @spread_id_2: 일치하는 바닥패 공간의 두번째 아이디
 * @multi_spread_id: 일치하는 바닥패에 두장이상인것중 첫번째의 공간아이디
 * 
 * 바닥패중 화투장 번호가 같은것들 다양한 정보를 파라미터로 넘어온 포인터에 넣어준다.
 */
static void
og_play_ui_spread_infos_by_card_number (gint     card_number,
                                        gint    *result_count,
                                        gint    *result_id_from_first_empty,
                                        gint    *result_spread_id_1,
                                        gint    *result_spread_id_2,
                                        gint    *result_single_spread_id,
                                        gint    *result_multi_spread_id)
{
    gint    match_count     = 0;
    gint    empty_id        = -1;
    gint    spread_ids[2];
    gint    single_grp_id   = -1;
    gint    multi_grp_id    = -1;

    gint i = 0;
    for (i = 0; i < SPRD_COUNT; i++)
    {
        CardSpread *card_group = &spreads[i];

        /* 조커 카운트를 빼는게 더 맞겠지만 아래 card_number검사를 하므로 별상관없다 */
        if (card_group->count > 0)
        {
            OgCard *card = og_cards[card_group->card_id[0]];

            if (card->number == card_number)
            {
                gint real_count = card_group->count - card_group->joker_count;

                spread_ids[match_count] = i;
                match_count++;

                if (real_count == 1)
                {
                    if (single_grp_id < 0)
                    {
                        single_grp_id = i;
                    }
                }
                else if (real_count > 1)
                {
                    if (multi_grp_id < 0)
                    {
                        multi_grp_id = i;
                    }
                }
            }
        }
        else if (empty_id < 0)
        {
            empty_id = i;
        }
        
        card_group = NULL;
    }

    if (result_id_from_first_empty)
    {
        *result_id_from_first_empty = empty_id;
    }
    
    if (result_count)
    {
        *result_count = match_count;
    }
    
    if (result_spread_id_1)
    {
        if (match_count >= 1)
        {
            *result_spread_id_1 = spread_ids[0];
        }
        else
        {
            *result_spread_id_1 = -1;
        }
            
    }
    
    if (result_spread_id_2)
    {
        if (match_count >= 2)
        {
            *result_spread_id_2 = spread_ids[1];
        }
        else
        {
            *result_spread_id_2 = -1;
        }
    }

    if (result_single_spread_id != NULL)
    {
        *result_single_spread_id = single_grp_id;
    }

    if (result_multi_spread_id != NULL)
    {
        *result_multi_spread_id = multi_grp_id;
    }
}

/*
 * og_play_ui_spread_add_card:
 * @spread_id: 바닥패 공간(그룹)의 index
 * @card_id: 전체 화투장중에서의 index
 * 
 * 바닥패 공간(그룹)에 화투장을 추가합니다. 화투장이 이미 존재하면 사선으로 약간 이동시켜 놓으며 없으면 빈공간에 놓습니다.
 */
static void
og_play_ui_spread_add_card (gint    spread_id,
                            gint    card_id)
{
    /* 바닥패 정보 추가 */
    CardSpread  *spread = &(spreads[spread_id]);
    OgCard      *card   = og_cards[card_id];

    spread->card_id[spread->count] = card_id;
    spread->count++;
    
    if (card->type == OG_CARD_TYPE_JOKER)
    {
        spread->joker_count++;
    }

    /* 화투장을 바닥패의 공간으로 이동 */
    gint    space   = ((spread->count - 1) * CARD_SPC_OVER);
    gint    x       = SPREAD_POSITION[spread_id].x + space;
    gint    y       = SPREAD_POSITION[spread_id].y + space;

    og_play_ui_card_move (og_cards[card_id], x, y, TRUE);
}

/*
 * og_play_ui_spread_merge_by_card_number:
 * @dest_spread_id: 도착지
 * @src_spread_id: 출발지
 * 
 * 바닥패를 합친다.
 */
static void
og_play_ui_spread_merge_by_card_number (gint dest_spread_id,
                                        gint src_spread_id)
{
    CardSpread *card_group = &(spreads[src_spread_id]);
    
    gint i = 0;
    for (i = 0; i < card_group->count; i++) {
        og_play_ui_spread_add_card (dest_spread_id, card_group->card_id[i]);
    }
    
    og_play_ui_card_spread_clear (&(spreads[src_spread_id]));
}

/*
 * og_play_ui_card_choice_show:
 * 
 * 바닥패 공간(그룹)중 자신이 낸 패와 같은 것이 2개 있을때 선택하는 화면을 나타낸다.
 */
static void
og_play_ui_card_choice_show (void)
{
    GtkWidget   *image_1    = og_play_ui_get_card_choice_image_1 ();
    GtkWidget   *image_2    = og_play_ui_get_card_choice_image_2 ();
    gint         card_id_1  = spreads[curr_submit_match_spread_ids[0]].card_id[0];
    gint         card_id_2  = spreads[curr_submit_match_spread_ids[1]].card_id[0];

    gtk_image_set_from_pixbuf (GTK_IMAGE (image_1), og_cards[card_id_1]->pixbuf);
    gtk_image_set_from_pixbuf (GTK_IMAGE (image_2), og_cards[card_id_2]->pixbuf);

    gtk_widget_show (og_play_ui_get_card_choice_window ());
}

/*
 * og_play_ui_card_choice_done:
 * 
 * 중복패 선택
 */
static void
og_play_ui_card_choice_done (gint choid_id)
{
    gtk_widget_hide (og_play_ui_get_card_choice_window ());
    og_play_ui_submit_done (curr_submit_match_spread_ids[choid_id],
            curr_submit_card_id);
}

/*
 * og_play_ui_submit:
 * @card_id: 전체 화투장중에서의 index
 * 
 * 플레이어의 패 또는 가운데 카드를 뒤집어 바닥패와 비교하여 상황에 맞게 처리한다.
 */
static void
og_play_ui_submit (gint card_id)
{
    OgCard  *card               = og_cards[card_id];
    gint     match_count        = 0;
    gint     dest_spread_id     = -1;
    gint     spread_id_1        = -1;
    gint     spread_id_2        = -1;
    gint     single_spread_id   = -1;
    gint     multi_spread_id    = -1;

    og_play_ui_player_submit_lock ();

    /* 이미 바닥에 같은패가 두개 깔려있을때 처음 나누어줄때는 세장을 겹쳐놓아야한다 */
    og_play_ui_spread_infos_by_card_number (card->number,
            &match_count, &dest_spread_id, &spread_id_1, &spread_id_2,
            &single_spread_id, &multi_spread_id);

    /* 바닥패중 현재패와 같은 번호가 두개인것이다. 이경우 4장인 화투는 절때 쌀수없으며 반드시 한장짜리가 존재한다. */
    if (match_count > 1)
    {
        /* 한공간에 두장이상 겹친 (조커 빼고) 공간이 하나도 없는 것이다. 둘중하나를 선택하게 해주어야한다. */
        if (multi_spread_id < 0)
        {
            curr_submit_match_spread_ids[0] = spread_id_1;
            curr_submit_match_spread_ids[1] = spread_id_2;
            curr_submit_card_id             = card_id;

            og_play_ui_card_choice_show ();

            return;
        }
        // 둘중 하나가 두장이 겹쳐있고 다른 하나가 한장이 경우이다.
        else
        {
            dest_spread_id = single_spread_id;
        }
    }
    /* 바닥패중 현재패와 같은 번호가 한개인경우 */
    else if (match_count == 1)
    {
        /* 일치하는 한공간이 두장이상 (조커 빼고)  겹쳐져 있는 경우. 쌌다 */
        if (multi_spread_id > -1)
        {
            dest_spread_id = multi_spread_id;
        }
        /* 일치하는 한공간이 한장만 있는경우 그곳에 넣으면 된다 */
        else
        {
            dest_spread_id = single_spread_id;
        }
    }
    /* 그밖의 경우에는 그냥 버리면 된다. dest_spread_id가 현재 빈공간의 아이디를 받은 상태이므로 그냥 사용하자 */

    og_play_ui_submit_done (dest_spread_id, card_id);
}

/*
 * og_play_ui_submit_done:
 * @spread_id: 바닥패 공간 인덱스
 * @card_id: 전체 화투장중에서의 인덱스
 * 
 * 플레이어의 패 또는 가운데 카드를 뒤집기 완료
 */
static void
og_play_ui_submit_done (gint    spread_id,
                        gint    card_id)
{
    gint match_id = curr_submit_type == SUBMIT_TYPE_PLAYER ? 0 : 1;

    /* 현재 제출한 바닥패 인덱스를 글로벌 변수 curr_match_spread_ids에 저장해둔다 */
    curr_match_spread_ids[match_id] = spread_id;

    /* 화투장을 바닥공간에 추가한다. */
    og_play_ui_spread_add_card (spread_id, card_id);

    if (curr_submit_type == SUBMIT_TYPE_PLAYER)
    {
        og_play_ui_open_card_by_player ();
        return;
    }
    else
    {
        /* 자기가 먹을것을 가져간다 */
        og_play_ui_submit_pickup (curr_player_id);
    }
}

static void
og_play_ui_submit_pickup (gint player_id)
{
    gint i = 0;
    gint j = 0;
    for (i = 0; i < G_N_ELEMENTS (curr_match_spread_ids); i++)
    {
        if (curr_match_spread_ids[i] < 0)
        {
            continue;
        }

        /*
         * 1. 본인이 내거나 뒤집은 화투장이 있는 공간만을 확인하는 루프이다.
         * 2. 조커외에 세장이면 싼거다. (1번의 조건에 따라)
         * 3. 반드시 조커이외의 화투장이 한장이상 있는 상태이다.
         * 4. 바닥에 조커가 있는 경우는 뒤집은 경우 뿐임.
         *          (자기것은 그냥 바로 점수패 공간으로 넘긴다.
         *          조커를 뒤집으면 자기가 낸 패 공간에 넣으므로 별도의 바닥 공간을 사용하지 않는다.)
         */

        /* 이 아래는 화투장이 있을때 처리 된다. 실제 패 또는 조커 */
        CardSpread  *spread         = &spreads[curr_match_spread_ids[i]];
        gint         real_count     = spread->count - spread->joker_count;
        gboolean     pick_normal    = FALSE;
        gboolean     pick_other     = FALSE;

        switch (real_count)
        {
            /* 조커만 먹는다 */
            case 1:
                if (spread->joker_count < 1)
                {
                    continue;
                }
                break;

            /* 다 먹는다 */
            case 2:
                pick_normal = TRUE;
                break;

            /* 쌌다 못먹는다. 이경우 조커도 그냥 같이 덮는다. 먹은 사람이 다가져간다. 길상동네 룰 */
            case 3:
                continue;
                break;

            /* 다 먹고 상대편 점수패중 피(껍데기)를 하나 가져온다. 1피부터 가져오고 없으면 쌍피 */
            case 4:
                pick_normal = TRUE;
                pick_other  = TRUE;
                break;
        }

        /* 뭔가 먹을게 있을때만 이리로 온다. 없으면 위에서 걸러진다 */

        /* 일단 자기가 먹을것을 다 자기것으로 옮긴다. */
        for (j = 0; j < spread->count; j++)
        {
            gint     card_id        = spread->card_id[j];
            OgCard  *card           = og_cards[card_id];

            if (pick_normal == FALSE && card->type != OG_CARD_TYPE_JOKER)
            {
                continue;
            }
            
            og_play_ui_player_add_point_card (player_id, card_id);

            /* 현재의 점수를 계산하여 유효한 (나는) 점수인경우 고할것인지 스톱할것인지를 띄운다 */
            CardPoint *point = &(player_point[player_id]);
            gint total_point = og_play_ui_card_point_get_total_point (point);
            if (total_point >= VALID_POINT && total_point > point->go_point) {
                gtk_widget_show (og_play_ui_get_go_choice_window ());
            }

            // TODO : 점수계산 그냥 찍어보는것임 나중에 지우자
            g_printf ("%d: %d\n", player_id, og_play_ui_card_point_get_total_point (&(player_point[player_id])));
        }
        og_play_ui_card_spread_clear (spread);

        /* 바닥패 쓸이의 경우 상대편 점수패중 피(껍데기)를 가져온다. 1피부터 가져오고 없으면 쌍피 */
        if (og_play_ui_get_spreads_using_count () < 1)
        {
            pick_other = TRUE;
        }

        if (pick_other)
        {
            // TODO 상대편 피(껍데기) 가져오는걸 구현한다.
        }
    }

    /* 플레이어 커서를 다음 사람으로 옮긴다 */
    og_play_ui_next_player_id ();

    og_play_ui_player_submit_unlock ();

    /* 네트웍 플레이가 아닐경우 다른 사용자의 플레이를 진행시킨다. */
    if (!network_play && curr_player_id != MY_PLAYER_ID)
    {
        /* TODO 테스트를 위해 마지막 카드를 낸다. 나중에는 먹은것과 낼 화투패를 참고하여 우선순위 알고리즘을 만들자 */
        gint hidden_id = og_play_ui_card_player_get_curr_id (&(player_cards[curr_player_id]));
        og_play_ui_player_submit_card (curr_player_id, player_cards[curr_player_id].hidden_card_ids[hidden_id]);
    }
}

/*
 * og_play_ui_player_submit_card:
 * @player_id: 플레이어의 index 상단 0, 하단 1
 * @card_id: 전체 화투장중에서의 index
 * 
 * 플레이어의 숨은 화투패중(hidden_card_ids[]) hidden_id에 해당하는 화투패를 제출한다.
 */
static void
og_play_ui_player_submit_card (gint player_id,
                               gint card_id)
{
    OgCard      *card   = og_cards[card_id];
    CardPlayer  *player = &(player_cards[player_id]);

    /* 매 턴마다 사용하는 변수들을 초기화한다 */
    og_play_ui_curr_match_spread_ids_clear ();
    og_play_ui_curr_submit_clear();

    curr_submit_type = SUBMIT_TYPE_PLAYER;

    if (player_id != MY_PLAYER_ID)
    {
        gint empty_id = player->empty_card_ids[og_play_ui_card_player_get_curr_id (player)];
        gtk_widget_hide (GTK_WIDGET (og_cards_empty[empty_id]));
    }
    else
    {
        og_play_ui_card_click_off (card);
    }

    og_play_ui_submit (card_id);
    player->submit_cnt++;
}

/*
 * og_play_ui_player_add_hidden_card:
 * @player_id: 플레이어의 index 상단 0, 하단 1
 * 
 * 플레이어의 숨은 화투패를 추가한다. 가운데 놓여진 뒤집어놓은 화투장에서 한장을 가져온다.
 */
static void
og_play_ui_player_add_hidden_card (gint player_id)
{
    OgCard      *og_card    = og_cards[curr_top_card_id];
    CardPlayer  *player     = &(player_cards[player_id]);

    gint i = 0;
    for (i = 0; i < PLAYER_CARD_COUNT; i++)
    {
        /* 가장 앞쪽에서부터 남는공간을 찾는다. */
        if (player->hidden_card_ids[i] == -1)
        {
            OgCard  **cards             = og_cards;
            gint      mv_card_id        = curr_top_card_id;
            gint      empty_top_card_id = og_play_ui_get_empty_top_card_id ();

            if (player_id == MY_PLAYER_ID)
            {
                og_play_ui_card_click_on (og_card);
                gtk_widget_hide (GTK_WIDGET (og_cards_empty[empty_top_card_id]));
            }
            else
            {
                cards       = og_cards_empty;
                mv_card_id  = empty_top_card_id; /* 빈 화투장 위젯을 상대방 패에 넣어둔다 */
            }

            /* 
             * 네트웍 플레이인경우 섞여진 실제 게임용 화투가 아닌 빈껍대기 화투장 위젯정보를
             * 사용자 화투장 정보에 넣어두어야한다. 실제정보는 실시간으로 서버에서 받아서 처리한다.
             * 혼자할때는 소스코드를 고쳐서 상대패를 보던 말던 별루 신경쓸필요 없다.
             */
            player->empty_card_ids[i] = empty_top_card_id;
            if (!network_play)
            {
                player->hidden_card_ids[i] = curr_top_card_id;
            }

            og_play_ui_card_move (cards[mv_card_id],
                    CARD_PSTN_PLYR_HIDN_X(player_id, i),
                    CARD_PSTN_PLYR_HIDN_Y(player_id, i / CARD_COUNT_PLYR_BY_LINE),
                    FALSE);

            og_play_ui_next_top_card_by_mixed ();

            break;
        }
    }
}

static void
og_play_ui_player_add_point_card (gint  player_id,
                                  gint  card_id)
{
    CardPoint   *point  = &(player_point[player_id]);
    OgCard      *card   = og_cards[card_id];
    gint         x      = -1;
    gint         y      = -1;

    switch (card->type)
    {
        case OG_CARD_TYPE_LIGHT:
            x = CARD_PSTN_PLYR_PNT_LGHT_X (player_id, point->light_cnt);
            y = CARD_PSTN_PLYR_PNT_LGHT_Y (player_id);
            point->light_card_ids[point->light_cnt] = card_id;
            point->light_cnt++;
            break;

        case OG_CARD_TYPE_TEN:
            x = CARD_PSTN_PLYR_PNT_TEN_X (player_id, point->ten_cnt);
            y = CARD_PSTN_PLYR_PNT_TEN_Y (player_id);
            point->ten_card_ids[point->ten_cnt] = card_id;
            point->ten_cnt++;
            if (card->sub_type == OG_CARD_SUB_TYPE_BIRD)
            {
                point->bird_cnt++;
            }
            break;

        case OG_CARD_TYPE_BELT:
            x = CARD_PSTN_PLYR_PNT_BLT_X (player_id, point->belt_cnt);
            y = CARD_PSTN_PLYR_PNT_BLT_Y (player_id);
            point->belt_card_ids[point->belt_cnt] = card_id;
            point->belt_cnt++;
            if (card->sub_type == OG_CARD_SUB_TYPE_BLUE)
            {
                point->blue_cnt++;
            }
            else if (card->sub_type == OG_CARD_SUB_TYPE_RED)
            {
                point->red_cnt++;
            }
            else if (card->sub_type == OG_CARD_SUB_TYPE_GRASS)
            {
                point->grass_cnt++;
            }
            break;

        case OG_CARD_TYPE_COVER:
            x = CARD_PSTN_PLYR_PNT_CVR_X (player_id, point->cover_cnt);
            y = CARD_PSTN_PLYR_PNT_CVR_Y (player_id, point->cover_cnt);
            point->cover_card_ids[point->cover_cnt] = card_id;
            point->cover_cnt++;
            if (card->sub_type == OG_CARD_SUB_TYPE_DOUBLE)
            {
                point->double_cnt++;
            }
            else if (card->sub_type == OG_CARD_SUB_TYPE_TRIPLE)
            {
                point->triple_cnt++;
            }
            break;

        // TODO 조커등을 구현
        case OG_CARD_TYPE_JOKER:
            break;
    }

    og_play_ui_card_move (og_cards[card_id], x, y, TRUE);    
}

/*
 * og_play_ui_open_card_by_player:
 * 
 * 플레이어가 가운데 뒤집어진 화투장중 하나를 연다.
 */
static void
og_play_ui_open_card_by_player (void)
{
    gint top_card_id = curr_top_card_id;

    og_play_ui_curr_submit_clear();

    /* 가운데 패 최상단 패를 안보이게 한다 */
    gtk_widget_hide (GTK_WIDGET (og_cards_empty[og_play_ui_get_empty_top_card_id ()]));
    /* 가운데 패 최상단 패의 커서를 그다음 장으로 옮긴다 */
    og_play_ui_next_top_card_by_mixed ();

    curr_submit_type = SUBMIT_TYPE_OPEN;
    og_play_ui_submit (top_card_id);
}

/*
 * og_play_ui_open_card_for_spread:
 * 
 * 가운데 뒤집어진 화투장중 한장을 열어 바닥 공간중 빈곳에 넣는다. 처음 화투장을 나누어 줄때 사용.
 */
static void
og_play_ui_open_card_for_spread (void)
{
    OgCard  *top_card           = og_play_ui_get_top_card ();
    gint     match_count        = 0;
    gint     dest_spread_id     = -1;
    gint     spread_id_1        = -1;
    gint     spread_id_2        = -1;
    gint     multi_spread_id    = -1;

    /* 이미 바닥에 같은패가 두개 깔려있을때 처음 나누어줄때는 세장을 겹쳐놓아야한다 */
    og_play_ui_spread_infos_by_card_number (top_card->number,
            &match_count, &dest_spread_id, &spread_id_1, &spread_id_2,
            NULL, &multi_spread_id);

    if (match_count > 1) {
        /* 같은패 두장을 합친다. */
        og_play_ui_spread_merge_by_card_number (spread_id_1,
                spread_id_2);

        dest_spread_id = spread_id_1;
    }
    else
    {
        /* 바닥에 깔린패중 두장이상이 겹쳐져있는경우 패를 나눠주는 현 시점에서는 세장을 겹쳐놓는다. */
        if (multi_spread_id > -1)
        {
            dest_spread_id = multi_spread_id;
        }
    }

    og_play_ui_spread_add_card (dest_spread_id, curr_top_card_id);

    gtk_widget_hide (GTK_WIDGET (og_cards_empty[og_play_ui_get_empty_top_card_id ()]));

    og_play_ui_next_top_card_by_mixed ();
}

/*
 * og_play_distribute_timeout:
 * @data: 사용자 정의 데이타 (현재 사용하지 않음)
 * 
 * 라운드 시작시 화투장을 바닥에 깔고 사용자에게 나누어준다. 약간의 딜레이를 주기 위해 타이머를 사용했다.
 * 
 * Return value: TRUE이면 타이머가 계속 돌고 FALSE이면 타이머는 멈춘다.
 */
static gboolean
og_play_distribute_timeout (gpointer data)
{
    gint i = 0;

    switch (dst_step)
    {
        case DST_STEP_NONE:
            /* 바닥 4장 (1) */
            for (i = 0; i < 4; i++)
            {
                og_play_ui_open_card_for_spread ();
            }

            dst_step = DST_STEP_SPRD_1;
            break;

        case DST_STEP_SPRD_1:
            /* 후 카드 1~5 */
            for (i = 0; i < 5; i++) {
                og_play_ui_player_add_hidden_card (curr_player_id);
            }
            og_play_ui_next_player_id ();
            
            dst_step = DST_STEP_PLYR_2_1;
            break;

        case DST_STEP_PLYR_2_1:
            /* 선 카드 1~5 */
            for (i = 0; i < 5; i++) {
                og_play_ui_player_add_hidden_card (curr_player_id);
            }
            og_play_ui_next_player_id ();

            dst_step = DST_STEP_PLYR_1_1;
            break;

        case DST_STEP_PLYR_1_1:
            /* 바닥 4장 (2) */
            for (i = 0; i < 4; i++)
            {
                og_play_ui_open_card_for_spread ();
            }

            dst_step = DST_STEP_SPRD_2;
            break;

        case DST_STEP_SPRD_2:
            /* 후 카드 6~10 */
            for (i = 5; i < 10; i++) {
                og_play_ui_player_add_hidden_card (curr_player_id);
            }
            og_play_ui_next_player_id ();

            dst_step = DST_STEP_PLYR_2_2;
            break;

        case DST_STEP_PLYR_2_2:
            /* 후 카드 6~10 */
            for (i = 5; i < 10; i++) {
                og_play_ui_player_add_hidden_card (curr_player_id);
            }
            og_play_ui_next_player_id ();

            dst_step = DST_STEP_PLYR_1_2;
            break;

        case DST_STEP_PLYR_1_2:
            og_play_ui_player_submit_unlock ();
            return FALSE;
            break;

        default:
            return FALSE;
            break;
    }

    return TRUE;
}

/*
 * og_play_ui_distribute_init:
 * 
 * 라운드 시작시 화투장을 배분한다.
 */
static void
og_play_ui_distribute_init (void)
{
    /* 바닥패 초기화 */
    og_play_ui_card_spreads_clear (spreads, SPRD_COUNT);
    og_play_ui_card_player_clear (&player_cards[0]);
    og_play_ui_card_player_clear (&player_cards[1]);
    og_play_ui_card_point_clear (&player_point[0]);
    og_play_ui_card_point_clear (&player_point[1]);
    dst_timeout_handler = g_timeout_add (DST_STEP_TIME, og_play_distribute_timeout, NULL);
}

/*
 * og_play_ui_remain_init:
 * 
 * 카드 전체를 사선으로 쌓아 가운데에 놓는다.
 */
static void
og_play_ui_remain_init (void)
{
    GtkWidget *fixed = og_play_ui_get_fixed ();

    gint i = 0;  
    for (i = 0; i < CARD_COUNT; i++)
    {
        gint x = CARD_PSTN_REMAIN_X + (i / 2);
        gint y = CARD_PSTN_REMAIN_Y + (i / 2);

        gtk_fixed_put (GTK_FIXED (fixed), GTK_WIDGET (og_cards[i]), 0, 0);
        gtk_fixed_put (GTK_FIXED (fixed), GTK_WIDGET (og_cards_empty[i]), x, y);
    }
}

/*
 * og_play_ui_mix_card:
 * 
 * 전체 화투장을 섞는다.
 */
static void
og_play_ui_mix_card (void)
{
    gint i = 0;
    for (i = 0; i < CARD_COUNT; i++)
    {
        og_cards[i]->mixed = FALSE;
        card_all_mixed_ids[i] = -1;
    }

    GRand *rand = g_rand_new ();
    for (i = 0; i < CARD_COUNT; i++)
    {
        gint range_end = (CARD_COUNT) - i;
        gint rand_id = g_rand_int_range (rand, 0, range_end);

        card_all_mixed_ids[i] = og_play_ui_get_relative_rand_card_id (rand_id);
        if (card_all_mixed_ids[i] == -1)
        {
            g_error ("mixed return value is -1");
        }
    }

    curr_top_card_id = card_all_mixed_ids[CARD_COUNT - 1];

    g_rand_free (rand);
    rand = NULL;
}

/*
 * public
 */

/**
 * og_play_ui_round_init:
 * 
 * 라운드 시작을 위해 화투장을 섞고 배분하는 등의 준비 작업을 한다.
 */
void
og_play_ui_round_init (void)
{
    if (!network_play)
    {
        og_play_ui_mix_card ();
    }
    og_play_ui_remain_init ();
    og_play_ui_distribute_init ();

    curr_player_id = player_first_id;
    og_play_ui_next_player_id ();
}

/*
 * og_play_ui_cover_pixbuf_new:
 * 
 * 화투장 커버 이미지(GdkPixbuf)를 화투장 크기에 맞게 새로 만들어 반환한다.
 * 
 * Return value: 새로 생성된 화투장 커버 이미지. 사용 완료시 g_object_unref 필요
 */
static GdkPixbuf*
og_play_ui_cover_pixbuf_new (void)
{
    GdkPixbuf   *pixbuf         = NULL;
    GdkPixbuf   *cover_pixbuf   = NULL;
    gchar       *filename       = NULL;
    
    filename = g_strdup_printf ("%s/image/card/hide.jpg", og_conf_get_exec_abs_path ());
    
    pixbuf = gdk_pixbuf_new_from_file (filename, NULL);
    cover_pixbuf = gdk_pixbuf_scale_simple (pixbuf, CARD_NML_WIDTH, CARD_NML_HEIGHT,
            GDK_INTERP_BILINEAR);

    filename = ot_sys_g_free (filename);
    
    g_object_unref (pixbuf);
    pixbuf = NULL;
    
    return cover_pixbuf;
}

/**
 * og_play_ui_init:
 * 
 * 사용자 UI를 초기화한다. 프로그램 실행시 한번만 한다. 이것은 반드시 싱글톤이 보장되어야한다,
 */
void
og_play_ui_init (void)
{
    if (init == FALSE)
    {
        gchar       *filename       = NULL;
        GdkPixbuf   *cover_pixbuf   = og_play_ui_cover_pixbuf_new ();

        og_cards        = g_malloc0 (sizeof (OgCard *) * CARD_COUNT);
        og_cards_empty  = g_malloc0 (sizeof (OgCard *) * CARD_COUNT);

        gint i = 0;
        for (i = 0; i < CARD_COUNT; i++)
        {
            OgCard  *og_card    = NULL;
            gint     num        = 0;
            gint     sub_num    = 0;

            num = (i / 4) + 1;
            sub_num = (i % 4) + 1;

            filename = g_strdup_printf ("%s/image/card/%02d_%02d.jpg",
                    og_conf_get_exec_abs_path (), num, sub_num);

            og_card = og_card_new_from_file (filename);
            og_cards[i] = og_card;

            og_card_set_number (og_card, num);
            og_card_set_type (og_card, CARD_TYPES[i].type, CARD_TYPES[i].sub_type);
            og_card_resize (og_card, CARD_NML_WIDTH, CARD_NML_HEIGHT);

            gtk_widget_hide (GTK_WIDGET (og_card));

            // finalize
            filename = ot_sys_g_free (filename);

            /*
             * 숨겨진 카드는 네트웍 플레이를 고려해 실제 카드정보를 가지고 있으면 안된다.
             * 빈카드를 실제 카드만큼 만들어 대체하자.
             * 정보들을 이용해 매번 이미지를 바꾸거나 하는 방법으로 하여 메모리를 절약할 수도 있겠지만
             * 메모리 몇메가 더 먹는것에 비해 코딩량도 많아지고 다른 리소스 사용량이 많아질것이다.
             */
            og_card = og_card_new ();
            og_cards_empty[i] = og_card;
            og_card_resize (og_card, CARD_NML_WIDTH, CARD_NML_HEIGHT);
            og_card_set_cover_pixbuf (og_card, cover_pixbuf);
            og_card_cover_show (og_card);

            gtk_widget_show (GTK_WIDGET (og_card));

            og_card = NULL;
        }

        if (cover_pixbuf != NULL) {
            g_object_unref (cover_pixbuf);
            cover_pixbuf = NULL;
        }


        /*
         * OgButton을 이용한 이미지 버튼 셋팅
         */

        GtkWidget   *go_choice_fixed    = og_play_ui_get_go_choice_fixed();
        OgButton    *button             = NULL;

        /* 고, 스톱 선택창 고버튼 */
        filename = g_strdup_printf ("%s/image/button/go.png", og_conf_get_exec_abs_path ());
        button = og_button_new_from_file (filename, 90, 50);
        gtk_fixed_put (GTK_FIXED (go_choice_fixed), GTK_WIDGET (button), 30, 170);
        filename = ot_sys_g_free (filename);

        /* 고, 스톱 선택창 스톱버튼 */
        filename = g_strdup_printf ("%s/image/button/stop.png", og_conf_get_exec_abs_path ());
        button = og_button_new_from_file (filename, 90, 50);
        gtk_fixed_put (GTK_FIXED (go_choice_fixed), GTK_WIDGET (button), 180, 170);
        filename = ot_sys_g_free (filename);


        init = TRUE;
    }
}

/**
 * og_play_ui_get_window:
 * 
 * 게임화면 Window 위젯을 반환한다.
 * 
 * Return value: 게임화면 Window 위젯
 */
GtkWidget*
og_play_ui_get_window (void)
{
    return glade_xml_get_widget (og_play_ui_get_glade_xml(), "og_play_window");
}

// TODO 삭제해주자.
void
og_ui_test (void)
{
    og_play_ui_round_init ();
}
