/* 화투장 위젯 소스

   This file is part of otak-godori.

   Copyright (C) 2007 Otak-godori Project.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  
*/

#include <gtk/gtk.h>

#include "../otak-lib/otak-util/otgtk.h"
#include "ogcard.h"


static GtkEventBoxClass *og_card_parent_class = NULL;



/*
 * signal
 */
gboolean
og_card_enter_notify_event_cb (GtkWidget        *widget,
                               GdkEventButton   *event,
                               gpointer         user_data)
{
    OgCard *card = OG_CARD (widget);

    if (card->highlight_cb != NULL) {
        card->highlight_cb (card);
    }

    gtk_widget_show (card->highlight_darea);

    return FALSE;
}

static gboolean
on_card_highlight_expose_event_cb (GtkWidget        *widget,
                                   GdkEventExpose   *event,
                                   gpointer         user_data)
{
    GtkWidget   *darea      = widget;
    GdkColor     white;
    GdkColor     black;
    GdkColor     selected;
    GdkGC       *gc         = NULL;
    GdkGC       *maskgc     = NULL;
    GdkPixmap   *mask       = NULL;
    gint         width      = darea->allocation.width;
    gint         height     = darea->allocation.height;

    gdk_color_parse ("black", &black);
    gdk_color_alloc (gtk_widget_get_colormap (darea), &black);

    gdk_color_parse ("white", &white);
    gdk_color_alloc (gtk_widget_get_colormap (darea), &white);

    gdk_color_parse ("blue", &selected);
    gdk_color_alloc (gtk_widget_get_colormap (darea), &selected);
    
    mask = gdk_pixmap_new (darea->window, width, height, 1);

    maskgc  = gdk_gc_new (mask);
    gc      = gdk_gc_new (darea->window);

    gdk_gc_set_foreground (gc, &selected);
    gdk_draw_rectangle (darea->window, gc, TRUE, 0, 0, width, height);
    
    gdk_gc_set_foreground (maskgc, &white);
    gdk_draw_rectangle (mask, maskgc, TRUE, 0, 0, width, height);

    gdk_gc_set_foreground (maskgc, &black);
    gdk_draw_rectangle (mask, maskgc, TRUE, 3, 3, width - 6, height - 6);

    gtk_widget_shape_combine_mask (darea, mask, 0, 0);

    g_object_unref (gc);
    g_object_unref (maskgc);
    g_object_unref (mask);

    return FALSE;
}

static gboolean
on_card_cover_expose_event_cb (GtkWidget        *widget,
                                   GdkEventExpose   *event,
                                   gpointer         user_data)
{
    OgCard *card = OG_CARD (user_data);

    if (card->cover_pixbuf != NULL) {
        gdk_draw_pixbuf (widget->window,
                widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                card->cover_pixbuf, 0, 0, 0, 0,
                widget->allocation.width, widget->allocation.height,
                GDK_RGB_DITHER_NORMAL, 0, 0);
    }

    return FALSE;
}

/*
 * private
 */

static void
og_card_destroy (GtkObject *object)
{
    OgCard *card = OG_CARD (object);

    if (card->image_widget != NULL)
    {
        gtk_widget_destroy (card->image_widget);
        card->image_widget = NULL;
    }

    if (card->highlight_darea != NULL)
    {
        gtk_widget_destroy (card->highlight_darea);
        card->highlight_darea = NULL;
    }

    if (card->cover_darea != NULL)
    {
        gtk_widget_destroy (card->cover_darea);
        card->cover_darea = NULL;
    }

    if (card->fixed != NULL)
    {
        gtk_widget_destroy (card->fixed);
        card->fixed = NULL;
    }

    if (card->pixbuf != NULL)
    {
        g_object_unref (card->pixbuf);
        card->pixbuf = NULL;
    }

    if (card->cover_pixbuf != NULL)
    {
        g_object_unref (card->cover_pixbuf);
        card->cover_pixbuf = NULL;
    }

    (* GTK_OBJECT_CLASS (og_card_parent_class)->destroy) (object);
}

/* class init */
static void
og_card_class_init (OgCardClass *klass)
{

    GtkObjectClass *object_class = GTK_OBJECT_CLASS (klass);

    og_card_parent_class = g_type_class_peek_parent (klass);

    // 이곳에 연결하면 gtk_widget_destroy시 함께 실행된다.
    object_class->destroy = og_card_destroy;
}

/* instance init */
static void
og_card_init (OgCard *card)
{
    card->number        = -1;
    card->type          = OG_CARD_TYPE_COVER;
    card->sub_type      = -1;
    card->mixed         = FALSE;
    card->highlight_cb  = NULL;

    card->pixbuf        = NULL;
    card->cover_pixbuf  = NULL;

    card->fixed             = GTK_WIDGET (gtk_fixed_new ());
    card->image_widget      = GTK_WIDGET (gtk_image_new ());
    card->highlight_darea   = GTK_WIDGET (gtk_drawing_area_new ());
    card->cover_darea       = GTK_WIDGET (gtk_drawing_area_new ());

    gtk_fixed_put (GTK_FIXED (card->fixed), card->image_widget, 0, 0);
    gtk_fixed_put (GTK_FIXED (card->fixed), card->highlight_darea, 0, 0);
    gtk_fixed_put (GTK_FIXED (card->fixed), card->cover_darea, 0, 0);
    gtk_container_add (GTK_CONTAINER (card), card->fixed);
    
    gtk_widget_show_all (card->fixed);
    gtk_widget_hide (card->highlight_darea);

    g_signal_connect ((gpointer) card->highlight_darea, "expose_event",
            G_CALLBACK (on_card_highlight_expose_event_cb), (gpointer) card);
    g_signal_connect ((gpointer) card->cover_darea, "expose_event",
            G_CALLBACK (on_card_cover_expose_event_cb), (gpointer) card);

    og_card_cover_hide (card);
}


/* get type */
GType
og_card_get_type (void)
{

    static GType widget_type = 0;

    if (G_UNLIKELY (widget_type == 0))
    {
        const GTypeInfo widget_info =
        {
            sizeof (OgCardClass),
            NULL,       /* base_init */
            NULL,
            (GClassInitFunc) og_card_class_init,
            NULL,       /* class_finalize */
            NULL,       /* class_data */
            sizeof (OgCard),
            0,          /* n_preallocs */
            (GInstanceInitFunc) og_card_init,
            NULL,       /* value_table */
        };
        widget_type = g_type_register_static (GTK_TYPE_EVENT_BOX, "OgCard",
                &widget_info, 0);
    }

    return widget_type;
}



/*
 * public
 */

OgCard*
og_card_new (void)
{
    OgCard *card = OG_CARD (g_object_new (OG_TYPE_CARD, NULL));

    return card;
}

OgCard*
og_card_new_from_file (const gchar *filename)
{
    OgCard *card = og_card_new ();

    if (card != NULL)
    {
        og_card_set_from_file (card, filename);
    }

    return card;
}

OgCard*
og_card_free (OgCard *card)
{
    return OG_CARD (ot_gtk_widget_free (GTK_WIDGET (card)));
}

void
og_card_set_from_file (OgCard       *card,
                       const gchar  *filename)
{
    if (card->pixbuf != NULL) {
        g_object_unref (card->pixbuf);
        card->pixbuf = NULL;
    }

    card->pixbuf = gdk_pixbuf_new_from_file (filename, NULL);
    gtk_image_set_from_pixbuf (GTK_IMAGE (card->image_widget), card->pixbuf);
}

void
og_card_set_number (OgCard  *card,
                    gint     number)
{
    card->number = number;
}

void
og_card_set_type (OgCard        *card,
                  OgCardType     type,
                  OgCardSubType  sub_type)
{
    card->type = type;
    card->sub_type = sub_type;
}

void
og_card_cover_show (OgCard *card)
{
    gtk_widget_show (card->cover_darea);

    if (GTK_WIDGET_REALIZED (card->cover_darea))
    {
        gdk_window_invalidate_rect (card->cover_darea->window, NULL, FALSE);
    }
}

void
og_card_cover_hide (OgCard *card)
{
    gtk_widget_hide (card->cover_darea);
}

void
og_card_set_cover_pixbuf (OgCard    *card,
                          GdkPixbuf *pixbuf)
{
    if (card->cover_pixbuf != NULL) {
        g_object_unref (card->cover_pixbuf);
        card->cover_pixbuf = NULL;
    }
    
    card->cover_pixbuf = pixbuf;
    g_object_ref (card->cover_pixbuf);
}

void
og_card_resize (OgCard  *card,
                gint    width,
                gint    height)
{
    g_return_if_fail (width > 0 && height > 0);

    gtk_widget_set_size_request (GTK_WIDGET (card), width, height);
    gtk_widget_set_size_request (GTK_WIDGET (card->image_widget), width, height);
    gtk_widget_set_size_request (GTK_WIDGET (card->highlight_darea), width, height);
    gtk_widget_set_size_request (GTK_WIDGET (card->cover_darea), width, height);
    gtk_widget_set_size_request (GTK_WIDGET (card->fixed), width, height);

    if (card->pixbuf != NULL)
    {
        GdkPixbuf *pixbuf_resize = gdk_pixbuf_scale_simple (card->pixbuf, width, height, GDK_INTERP_BILINEAR);
    
        gtk_image_set_from_pixbuf (GTK_IMAGE (card->image_widget), pixbuf_resize);
    
        g_object_unref (pixbuf_resize);
        pixbuf_resize = NULL;
    }
}

void
og_card_highlight_connect (OgCard           *card,
                           OgCardSimpleFunc highlight_cb)
{
    g_signal_connect ((gpointer) card, "enter_notify_event",
            G_CALLBACK (og_card_enter_notify_event_cb), NULL);
    card->highlight_cb = highlight_cb;
}

void
og_card_highlight_disconnect (OgCard *card)
{
    g_signal_handlers_disconnect_by_func ((gpointer) card,
            G_CALLBACK (og_card_enter_notify_event_cb), NULL);
}

void
og_card_highlight_off (OgCard *card)
{
    gtk_widget_hide (card->highlight_darea);
}
