/* 화투장 위젯 헤더

   This file is part of otak-godori.

   Copyright (C) 2007 Otak-godori Project.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
   Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  
*/

#ifndef OGCARD_H_
#define OGCARD_H_

#include <gtk/gtkeventbox.h>
#include <gtk/gtkimage.h>

G_BEGIN_DECLS

#define OG_TYPE_CARD            (og_card_get_type ())
#define OG_CARD(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), OG_TYPE_CARD, OgCard))
#define OG_CARD_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), OG_TYPE_CARD, OgCardClass))
#define OG_IS_CARD(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), OG_TYPE_CARD))
#define OG_IS_CARD_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), OG_TYPE_CARD))
#define OG_CARD_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), OG_TYPE_CARD, OgCardClass))


typedef struct _OgCard      OgCard;
typedef struct _OgCardClass OgCardClass;

typedef enum
{
    OG_CARD_TYPE_COVER, /* 피(껍데기) */
    OG_CARD_TYPE_BELT,  /* 띠 */
    OG_CARD_TYPE_LIGHT, /* 광 */
    OG_CARD_TYPE_TEN,   /* 열끗 */
    OG_CARD_TYPE_JOKER  /* 조커 */
} OgCardType; /* 기본 타입 */

typedef enum
{
    OG_CARD_SUB_TYPE_NONE,      /* 없음. BELT(띠)이고 NONE이면 초단 */
    OG_CARD_SUB_TYPE_BLUE,      /* 청단 */
    OG_CARD_SUB_TYPE_RED,       /* 홍단 */
    OG_CARD_SUB_TYPE_BIRD,      /* 고도리 */
    OG_CARD_SUB_TYPE_GRASS,     /* 초단 */
    OG_CARD_SUB_TYPE_DOUBLE,    /* 쌍피 */
    OG_CARD_SUB_TYPE_TRIPLE     /* 삼피 */
} OgCardSubType; /* 부 타입 */

typedef void (*OgCardSimpleFunc) (OgCard *card);

struct _OgCard
{
    GtkEventBox event_box;

    GtkWidget   *fixed;
    GtkWidget   *image_widget;
    GtkWidget   *highlight_darea;
    GtkWidget   *cover_darea;
    
    GdkPixbuf   *pixbuf;
    GdkPixbuf   *cover_pixbuf;

    gint            number;     /* 화투장 번호 */
    OgCardType      type;       /* 띠, 피(껍데기), 광 등등 */
    OgCardSubType   sub_type;   /* 부(상세)타입. 띠는 청단이냐 홍단이냐, 피(껍데기)는 쌍피냐 아니냐 */
    gboolean        mixed;      /* 화투를 섞을때 이미 배정되었는지 여부 */

    OgCardSimpleFunc highlight_cb;
};

struct _OgCardClass {
    GtkEventBoxClass parent_class;
};

GType   og_card_get_type                (void);
OgCard* og_card_new                     (void);
OgCard* og_card_new_from_file           (const gchar *filename);
OgCard* og_card_free                    (OgCard *card);

void    og_card_set_from_file           (OgCard         *card,
                                         const gchar    *filename);
void    og_card_set_number              (OgCard *card,
                                         gint   number);
void    og_card_set_type                (OgCard         *card,
                                         OgCardType      type,
                                         OgCardSubType   sub_type);
void    og_card_cover_show              (OgCard *card);
void    og_card_cover_hide              (OgCard *card);
void    og_card_set_cover_pixbuf        (OgCard     *card,
                                         GdkPixbuf  *pixbuf);

void    og_card_resize                  (OgCard *card,
                                         gint   width,
                                         gint   height);
void    og_card_highlight_connect       (OgCard             *card,
                                         OgCardSimpleFunc    highlight_cb);
void    og_card_highlight_disconnect    (OgCard *card);
void    og_card_highlight_off           (OgCard *card);

G_END_DECLS

#endif /*OGCARD_H_*/
